# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math

import networkx as nx

from slambuc.alg import INFEASIBLE, T_RESULTS, T_BARRS_GEN
from slambuc.alg.util import (ipowerset, isubtrees, par_subtree_memory, ibacktrack_chain, par_subtree_cost,
                              par_subchain_latency)


def isubtrees_exhaustive(tree: nx.DiGraph, root: int, M: int, N: int = 1) -> T_BARRS_GEN:
    """
    Calculate all combinations of edge cuts and returns only if it is feasible wrt. the memory limit *M*.

    Block metrics are calculated based on parallelized execution platform model.

    :param tree:    service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param root:    root node of the graph
    :param M:       upper memory bound in MB
    :param N:       available CPU core count
    :return:        generator of feasible subtrees' barrier nodes
    """
    for cuts in ipowerset(tree.edges(range(1, len(tree)))):
        barrs = {root}.union(v for _, v in cuts)
        # Check whether the subtrees meet the memory requirement M
        feasible_subtrees = [b for b, nodes in isubtrees(tree, barrs) if par_subtree_memory(tree, b, nodes, N) <= M]
        if len(feasible_subtrees) == len(barrs):
            yield feasible_subtrees


def greedy_par_tree_partitioning(tree: nx.DiGraph, root: int = 1, M: int = math.inf, L: int = math.inf,
                                 N: int = 1, cp_end: int = None, delay: int = 1) -> T_RESULTS:
    """
    Calculate minimal-cost partitioning of a service graph(tree) by greedily iterating over all possible cuttings.

    Block metrics are calculated based on parallelized execution platform model.

    :param tree:    service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param root:    root node of the graph
    :param M:       upper memory bound of the partition blocks (in MB)
    :param L:       latency limit defined on the critical path (in ms)
    :param N:       available CPU core count
    :param cp_end:  tail node of the critical path in the form of subchain[root -> c_pend]
    :param delay:   invocation delay between blocks
    :return:        tuple of list of best partitions, sum cost of the partitioning, and resulted latency
    """
    cp_end = cp_end if cp_end is not None else max(tree)
    best_res, best_cost = [INFEASIBLE], math.inf
    cpath = set(ibacktrack_chain(tree, root, cp_end))
    # Iterates over all possible cuttings
    for barrs in isubtrees_exhaustive(tree, root, M, N):
        partition = []
        sum_cost = 0
        for b, nodes in isubtrees(tree, barrs):
            partition.append(nodes)
            sum_cost += par_subtree_cost(tree, b, nodes, N)
        partition.sort()
        # Calculate blocks of critical path based on the partitioning
        restricted_blk_lats = [par_subchain_latency(tree, blk[0], set(blk), cpath, N) for blk in partition]
        sum_lat = sum(restricted_blk_lats) + (sum(map(bool, restricted_blk_lats)) - 1) * delay
        if sum_lat <= L:
            # Store partitioning with the same best cost for comparison
            if sum_cost == best_cost:
                best_res.append((partition, sum_cost, sum_lat))
            # Initialize new best cost partitioning
            elif sum_cost < best_cost:
                best_res, best_cost = [(partition, sum_cost, sum_lat)], sum_cost
    return best_res



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math
import random

from slambuc.alg.chain.dp.greedy import greedy_chain_partitioning
from slambuc.misc.util import evaluate_chain_partitioning


def run_test(runtime: list, memory: list, rate: list, M: int = math.inf, N: int = math.inf, L: int = math.inf,
             start: int = 0, end: int = None, delay: int = 1, unit: int = 100):
    results = greedy_chain_partitioning(runtime, memory, rate, M, N, L, start, end, delay, unit)
    for i, (partition, opt_cost, opt_lat) in enumerate(results):
        print(f"  GREEDY[{i}]  ".center(80, '#'))
        evaluate_chain_partitioning(partition, opt_cost, opt_lat, runtime, memory, rate, M, N,
                                    L, start, end, delay, unit)
    return results


def test_chain():
    runtime = [20, 40, 50, 20, 70, 40, 50, 60, 40, 10]
    memory = [3, 3, 2, 1, 2, 1, 2, 1, 2, 3]
    rate = [1, 1, 2, 2, 1, 3, 1, 2, 1, 3]
    delay = 10
    M = 6
    N = 3
    L = 500
    start = 0
    end = len(runtime) - 1
    unit = 100
    run_test(**locals())


def test_random_chain():
    runtime = [random.randint(10, 100) for _ in range(10)]
    memory = [random.randint(1, 3) for _ in range(10)]
    rate = [random.randint(1, 3) for _ in range(10)]
    delay = 10
    M = 6
    N = 2
    L = sum(runtime) + 1 + delay * random.randint(len(runtime) // 4, len(runtime) // 2)
    start = 0
    end = len(runtime) - 1
    unit = 100
    run_test(**locals())


if __name__ == '__main__':
    test_chain()
    # test_random_chain()



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math
import pathlib

import networkx as nx
import tabulate

from slambuc.alg.service import NAME
from slambuc.alg.tree.ser.pseudo import pseudo_btree_partitioning, pseudo_ltree_partitioning
from slambuc.alg.util import ipostorder_dfs, ileft_right_dfs
from slambuc.misc.generator import get_random_tree
from slambuc.misc.plot import draw_tree
from slambuc.misc.util import evaluate_ser_tree_partitioning


def test_btree_traversal(draw: bool = True):
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree_ser.gml", destringizer=int)
    tree.graph[NAME] += "-ser_pseudo"
    subcases = []
    for p, v in ipostorder_dfs(tree, 1):
        # print(f"({p=} -->)  [{v}]")
        subcases.append([f"{p}     -", f"[{v}]", 0])
        for i, b in enumerate(tree.succ[v], start=1):
            # print(f"({p=} -->)  [{v=}]  --({i=})-->  {b=}")
            subcases.append([f"{p}     -", f"[{v}]", f"--({i})-->", f"[{b}]"])
    print(tabulate.tabulate(subcases, headers=['p', 'v', 'i', 'b'], stralign='center', tablefmt='presto'))
    if draw:
        draw_tree(tree)


def test_ltree_traversal(draw: bool = True):
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree_ser.gml", destringizer=int)
    tree.graph[NAME] += "-ser_pseudo"
    subcases = []
    for p, n in ipostorder_dfs(tree, 1):
        subcases.append([f"T[{n}]"])
        for (u, v_p), b_p, (v, b) in ileft_right_dfs(tree, n):
            subcases.append([])
            subcases.append([None, None, "|-->", f"({v_p})" if v_p else None, "|-->",
                             f"({b_p})" if b_p else None])
            subcases.append([None, u if v != n else f"({p})", "-------->", v, "-------->", b])
    print(tabulate.tabulate(subcases, headers=['n', 'u', '', 'v', '', 'b'], stralign='right', tablefmt='presto'))
    print("Iteration for the whole tree:")
    for (u, v_p), b_p, (v, b) in ileft_right_dfs(tree, 1):
        print(u, v_p, '-', b_p, '-', v, b)
    if draw:
        draw_tree(tree)


def run_test(tree: nx.DiGraph, M: int, L: int, root: int = 1, cp_end: int = None, delay: int = 10):
    partition, opt_cost, opt_lat = pseudo_btree_partitioning(tree, root, M, L, cp_end, delay)
    evaluate_ser_tree_partitioning(tree, partition, opt_cost, opt_lat, root, cp_end, M, L, delay)
    partition, opt_cost, opt_lat = pseudo_ltree_partitioning(tree, root, M, L, cp_end, delay)
    evaluate_ser_tree_partitioning(tree, partition, opt_cost, opt_lat, root, cp_end, M, L, delay)
    return partition, opt_cost, opt_lat


def test_ser_tree_pseudo_partitioning():
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree_ser.gml", destringizer=int)
    tree.graph[NAME] += "-ser_pseudo"
    params = dict(tree=tree,
                  root=1,
                  cp_end=10,
                  M=6,
                  L=math.inf,
                  # L = 430,
                  delay=10)
    run_test(**params)


def test_random_tree_partitioning(n: int = 10):
    tree = get_random_tree(n)
    tree.graph[NAME] += "-ser_pseudo"
    params = dict(tree=tree,
                  root=1,
                  cp_end=n,
                  M=6,
                  L=math.inf,
                  # L = 430,
                  delay=10)
    run_test(**params)


if __name__ == '__main__':
    # test_btree_traversal()
    # test_ltree_traversal()
    test_ser_tree_pseudo_partitioning()
    # test_random_tree_partitioning()



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import importlib.resources
import itertools
import pathlib
import random
from collections.abc import Generator

import networkx as nx
import numpy as np
import scipy

from slambuc.alg.service import *
from slambuc.gen.io import load_hist_params, save_trees_to_file
from slambuc.gen.microservice.power_ba_graph import generate_power_ba_graph
from slambuc.misc.plot import draw_tree

# Distribution parameters of Serverless functions
# See also: https://dl.acm.org/doi/10.5555/3489146.3489160
# Avg. function execution time in sec ~ Log-Normal(mu, sigma)
RT_MU = -0.38
RT_SIGMA = 2.36
# Scale down execution time values (recorded in sec) one magnitude lower compared to microservice components
RT_DIST = scipy.stats.lognorm(s=RT_SIGMA, loc=0.0, scale=np.exp(RT_MU) * 100)

# Avg. function memory demand in MB ~ Burr(c, k, lambda)
MEM_C = 11.652
MEM_K = 0.221
MEM_LAMBDA = 107.083
MEM_DIST = scipy.stats.burr12(c=MEM_C, d=MEM_K, loc=0.0, scale=MEM_LAMBDA)

# Empirical distribution parameters of function calls extracted from Alibaba traces
# See also: https://github.com/alibaba/clusterdata/tree/master/cluster-trace-microservices-v2021
HIST_DIR = importlib.resources.files("slambuc.gen.microservice").joinpath("hist")
# Invocation rate per services
RATE_HIST_NAME = "func_inv_rate_hist"
RATE_DIST = scipy.stats.rv_histogram(load_hist_params(HIST_DIR, RATE_HIST_NAME), density=True)

# Data R/W overhead
DATA_HIST_NAME = "rw_overhead_hist"
DATA_DIST = scipy.stats.rv_histogram(load_hist_params(HIST_DIR, DATA_HIST_NAME), density=True)

# Tree structure parameters of preferential attachment (Alpha) and attractiveness of leafs (a)
PREF_ATT_LOW, PREF_ATT_HIGH = 0.05, 0.9
LEAF_ATTR_LOW, LEAF_ATTR_HIGH = 0.1, 3.25

DEF_FAAS_TREE_PREFIX = f"faas_tree"


def ifunc_attributes(n: int, dist: scipy.stats.rv_continuous, transform=np.round) -> Generator[int]:
    """
    Generate attribute values of the given size *n* base on the given distribution *dist*.

    :param n:           number of attributes
    :param dist:        build distribution object
    :param transform:   transform function applied on every attribute value
    :return:            generator of attributes
    """
    yield from transform(dist.rvs(size=n)).astype(int)


def get_faas_tree(n: int, Alpha: float = 1.0, a: float = 0.0) -> nx.DiGraph:
    """
    Generate service tree with attributes drawn from the predefined distributions.

    :param n:       number of nodes
    :param Alpha:   power of preferential attachment (default: 1.0)
    :param a:       attractiveness of vertices with no edges (default: 0.0)
    :return:        generated tree
    """
    tree = nx.bfs_tree(generate_power_ba_graph(n=n, m=1, Alpha=Alpha, a=a, root=1), source=1, sort_neighbors=sorted)
    runtimes, memories = ifunc_attributes(n, RT_DIST), ifunc_attributes(n, MEM_DIST)
    rates, data = ifunc_attributes(n, RATE_DIST, transform=np.floor), ifunc_attributes(n, DATA_DIST)
    tree.add_edge(PLATFORM, 1)
    for u, v in nx.dfs_edges(tree, source=PLATFORM):
        tree.nodes[v][RUNTIME], tree.nodes[v][MEMORY] = next(runtimes), next(memories)
        tree[u][v][RATE], tree[u][v][DATA] = next(rates), next(data)
    tree[PLATFORM][1][RATE] = 1
    tree.graph[NAME] = f"faas_tree_n{n}A{Alpha}a{a}"
    return tree


def verify_faas_tree(n: int = 10):
    """
    Plot random generated serverless tree.

    :param n:   tree size
    """
    t = get_faas_tree(n)
    draw_tree(t, draw_weights=True)


def generate_all_faas_trees(data_dir: str, Alpha: float = PREF_ATT_HIGH, a: float = LEAF_ATTR_HIGH,
                            iteration: int = 100, start: int = 10, end: int = 100, step: int = 10,
                            tree_name: str = DEF_FAAS_TREE_PREFIX):
    """
    Generate Serverless/Faas service trees with attributes from predefined and extracted distributions.

    :param data_dir:    directory of saved trees
    :param Alpha:       power of preferential attachment (default: 1.0)
    :param a:           attractiveness of vertices with no edges (default: 0.0)
    :param iteration:   number of generated trees
    :param start:       minimum of size intervals
    :param end:         maximum of size intervals
    :param step:        step size of intervals
    :param tree_name:   prefix name of tree files
    """
    for min_size, max_size in itertools.pairwise(range(start, end + step, step)):
        print(f"Generating Serverless/FaaS trees with {min_size} <= size <= {max_size}...")
        trees = [get_faas_tree(n=random.randint(min_size, max_size), Alpha=Alpha, a=a)
                 for _ in range(iteration)]
        file_name = pathlib.Path(data_dir, f"{tree_name}_n{min_size}-{max_size}.npy").resolve()
        print(f"Saving trees into {file_name}...")
        save_trees_to_file(trees, file_name, padding=max_size)


if __name__ == '__main__':
    verify_faas_tree(n=20)
    # generate_all_faas_trees("../../../validation/data")



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pathlib

import networkx as nx

from slambuc.alg.ext.min_cut import (min_weight_subchain_split, min_weight_chain_decomposition, min_weight_ksplit,
                                     min_weight_ksplit_clustering)
from slambuc.alg.util import recreate_subchain_blocks
from slambuc.misc.plot import draw_tree
from slambuc.misc.util import evaluate_par_tree_partitioning


def test_min_weight_chain_decomposition(root: int = 1, N: int = 1, cp_end: int = 10, delay: int = 10):
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree.gml", destringizer=int)
    barr = min_weight_subchain_split(tree, root)
    partition = recreate_subchain_blocks(tree, barr)
    print("Partition:", partition)
    partition, sum_cost, sum_lat = min_weight_chain_decomposition(tree, root, N, cp_end, delay)
    evaluate_par_tree_partitioning(tree, partition, sum_cost, sum_lat, root, cp_end, None, None, N, delay, draw=False)
    draw_tree(tree, partition, draw_blocks=True, draw_weights=True)


def test_min_weight_tree_partitioning(k: int = 2, root: int = 1, N: int = 1, cp_end: int = 10, delay: int = 10):
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree.gml", destringizer=int)
    barr = min_weight_ksplit(tree, root, k)
    partition = recreate_subchain_blocks(tree, barr)
    print("Partition", partition)
    partition, sum_cost, sum_lat = min_weight_ksplit_clustering(tree, root, k, N, cp_end, delay)
    evaluate_par_tree_partitioning(tree, partition, sum_cost, sum_lat, root, cp_end, None, None, N, delay, draw=False)
    draw_tree(tree, partition, draw_blocks=True, draw_weights=True)


if __name__ == '__main__':
    # test_min_weight_chain_decomposition()
    test_min_weight_tree_partitioning(k=3)



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import itertools

import networkx as nx

from slambuc.alg.service import PLATFORM, RUNTIME, DATA
from slambuc.alg.util import ipostorder_dfs


def faasify_dag_by_duplication(dag: nx.DiGraph, root: int) -> nx.DiGraph:
    """
    One-way transformation of a DAG of modules/components into a tree by iteratively duplicating sub-graphs
    related to nodes with multiple predecessors.

    The algorithm requires that the input DAG must have only one source node.

    :param dag:     input DAG
    :param root:    root node
    :return:        generated tree
    """
    if dag is None:
        return
    while not nx.is_tree(dag):
        new_id = itertools.count(max(filter(lambda _v: _v is not PLATFORM, dag.nodes)) + 1)
        for v in list(dag.nodes):
            if len(dag.pred[v]) > 1:
                while len(ingress := list(dag.pred[v])) > 1:
                    p = ingress.pop(0)
                    e_data = dag[p][v]
                    st = dag.subgraph(nx.dfs_postorder_nodes(dag, source=v))
                    relabel = {sv: next(new_id) for sv in st.nodes}
                    st = nx.relabel_nodes(st, mapping=relabel)
                    dag.update(st)
                    dag.remove_edge(p, v)
                    dag.add_edge(p, relabel[v], **e_data)
    # Add front-end dispatcher node for nodes without predecessors
    shift = max(filter(lambda _v: _v is not PLATFORM, dag.nodes))
    # Relabel nodes to confirm to our service tree structure demands
    shifted_mapper = {v: v + shift for _, v in nx.bfs_edges(dag, source=root)}
    nx.relabel_nodes(dag, mapping=shifted_mapper, copy=False)
    relabel = itertools.count(1)
    relabel_mapper = {v: next(relabel) for _, v in nx.bfs_edges(dag, source=PLATFORM, sort_neighbors=sorted)}
    nx.relabel_nodes(dag, mapping=relabel_mapper, copy=False)
    return dag


def faasify_dag_by_cutting(dag: nx.DiGraph, root: int) -> nx.DiGraph:
    ...
    # TODO


def transform_autonomous_caching(tree: nx.DiGraph, root: int, copy: bool = False) -> nx.DiGraph:
    """
    Transform given *tree* by adding fetching and out-caching overheads to function execution times.

    :param tree:    input tree
    :param root:    root node
    :param copy:    use a deep copy of the input instead of modifying the original
    :return:        transformed tree
    """
    tf_tree = tree.copy() if copy else tree
    for p, n in ipostorder_dfs(tree, root):
        # Add data fetching and state caching overheads to the function execution time
        tf_tree.nodes[n][RUNTIME] += tree[p][n][DATA] + sum(tree[n][s][DATA] for s in tree.successors(n))
    return tf_tree



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import itertools
import warnings

import networkx as nx
from matplotlib import pyplot as plt

from slambuc.alg.ext.csp import decode_state, START, END, COST, LAT
from slambuc.alg.service.common import *
from slambuc.alg.util import path_blocks

PART_COLORS = ('red', 'orange', "brown", 'green', "purple", "blue", "black", "magenta")


def draw_tree(tree: nx.DiGraph, partition: list = None, cuts: list = None, draw_weights=False, draw_blocks=False,
              figsize=None, ax=None, **kwargs):
    """
    Draw tree with given partitioning in a top-down topological structure.

    :param tree:            service tree
    :param partition:       calculated partitioning (optional)
    :param cuts:            calculated cuts (optional)
    :param draw_weights:    draw node/edge weights instead of IDs
    :param draw_blocks:     draw surrounding blocks
    :param figsize:         figure dimensions (optional)
    :param ax:              matplotlib axis (optional)
    """
    if figsize is None:
        d = nx.dag_longest_path_length(tree)
        figsize = (d, d)
    if ax is None:
        plt.figure(figsize=figsize, dpi=300)
        ax = plt.gca()
    colors = itertools.cycle(PART_COLORS)
    for v in tree.nodes:
        if COLOR in tree.nodes[v]:
            del tree.nodes[v][COLOR]
    tree.nodes[PLATFORM][COLOR] = "gray"
    if partition:
        if isinstance(partition[0], tuple):
            if len(PART_COLORS) < len(set(p[-1] for p in partition)):
                warnings.warn(f"Not enough colors({len(PART_COLORS)}) for all different flavor!")
            f_colors = dict()
            for (blk, f) in partition:
                if f not in f_colors:
                    f_colors[f] = next(colors)
                for n in blk:
                    tree.nodes[n][COLOR] = f_colors[f]
            partition = [blk[0] for blk in partition]
        else:
            for node, pred in nx.bfs_predecessors(tree, PLATFORM):
                if COLOR in tree.nodes[node]:
                    continue
                blk = 0
                while node not in partition[blk]:
                    blk += 1
                color = next(colors)
                if COLOR in tree.nodes[pred]:
                    while tree.nodes[pred][COLOR] == color:
                        color = next(colors)
                for n in partition[blk]:
                    tree.nodes[n][COLOR] = color
        node_colors = [tree.nodes[n][COLOR] for n in tree.nodes]
    else:
        node_colors = ["tab:gray" if n is PLATFORM else "tab:green" for n in tree.nodes]
    if cuts:
        edge_colors = ["tab:red" if e in cuts else "black" for e in tree.edges]
    else:
        edge_colors = "black"
    if draw_weights:
        labels = {n: f"T{tree.nodes[n][RUNTIME]}\nM{tree.nodes[n][MEMORY]}" for n in tree.nodes if n is not PLATFORM}
    else:
        labels = {n: n for n in tree.nodes}
    labels[PLATFORM] = PLATFORM
    pos = nx.drawing.nx_agraph.graphviz_layout(tree, prog='dot', root=0)
    nx.draw(tree, ax=ax, pos=pos, arrows=True, arrowsize=20, width=2, with_labels=True, node_size=1000, font_size=10,
            font_color="white", labels=labels, node_color=node_colors, edge_color=edge_colors, **kwargs)
    if draw_weights:
        e_labels = {(u, v): f"R{tree[u][v][RATE]}\nD{tree[u][v][DATA]}" for u, v in tree.edges}
        nx.draw_networkx_edge_labels(tree, pos, edge_labels=e_labels, label_pos=0.5, font_size=10)
    if draw_blocks:
        dist_x = sorted({k[0] for k in pos.values()})
        dist_y = sorted({k[1] for k in pos.values()})
        off_x = 0.5 * (sum(abs(b - a) for a, b in itertools.pairwise(dist_x)) // len(dist_x)
                       if len(dist_x) > 1 else pos[PLATFORM][0])
        off_y = 0.5 * (sum(abs(b - a) for a, b in itertools.pairwise(dist_y)) // len(dist_y)
                       if len(dist_y) > 1 else pos[PLATFORM][1])
        for blk in partition:
            lefts, rights = ([(pos[blk[0]][0] - off_x, pos[blk[0]][1] + off_y)],
                             [(pos[blk[0]][0] + off_x, pos[blk[0]][1] + off_y)])
            levels = path_blocks(list(nx.topological_generations(tree)), blk)
            for i, lvl in enumerate(levels):
                lefts.append((pos[min(lvl)][0] - off_x, pos[min(lvl)][1]))
                rights.append((pos[max(lvl)][0] + off_x, pos[max(lvl)][1]))
            lefts.append((pos[min(levels[-1])][0] - off_x, pos[min(levels[-1])][1] - off_y))
            rights.append((pos[max(levels[-1])][0] + off_x, pos[max(levels[-1])][1] - off_y))
            rights.reverse()
            poly = plt.Polygon(lefts + rights, closed=True, fc=tree.nodes[blk[0]][COLOR], ec=tree.nodes[blk[0]][COLOR],
                               lw=3, ls=':', fill=True, alpha=0.3, capstyle='round', zorder=0)
            ax.add_patch(poly)
    plt.title(tree.graph[NAME])
    plt.tight_layout()
    plt.show()
    plt.close()


def draw_state_dag(dag: nx.DiGraph, chains: list[list[int]], draw_weights: bool = False):
    """
    Draw state-space DAG in a vertically-ordered multipartite layout.

    :param dag:             input DAG
    :param chains:          chain decomposition of the given tree
    :param draw_weights:    draw node/edge weights instead of IDs
    """
    h = max(len(c) for c in chains) * 1.5
    plt.figure(figsize=(2 * h, h), dpi=300)
    layers = [START, *itertools.chain.from_iterable(chains), END]
    for v in dag.nodes:
        dag.nodes[v]['layer'] = layers.index(decode_state(v)[0][0] if v not in (START, END) else v)
    node_color = ["tab:green" if n is START else "tab:red" if n is END else "tab:blue" for n in dag.nodes]
    pos = nx.multipartite_layout(dag, subset_key='layer', scale=1)
    nx.draw(dag, pos=pos, arrows=True, arrowsize=15, width=2, with_labels=True, node_size=1000, font_size=10,
            node_color=node_color)
    if draw_weights:
        e_labels = {(u, v): f"C:{dag[u][v][COST]},L:{dag[u][v][LAT][1]}" for u, v in dag.edges}
        nx.draw_networkx_edge_labels(dag, pos, edge_labels=e_labels, label_pos=0.5, font_size=10)
    plt.show()
    plt.close()



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math

from slambuc.alg.chain.ser.greedy import greedy_ser_chain_partitioning
from slambuc.misc.generator import get_random_chain_data
from slambuc.misc.util import evaluate_ser_chain_partitioning, print_ser_chain_summary


def run_test(runtime: list, memory: list, rate: list, data: list, M: int = math.inf, L: int = math.inf, start: int = 0,
             end: int = None, delay: int = 1):
    results = greedy_ser_chain_partitioning(runtime, memory, rate, data, M, L, start, end, delay)
    for i, (partition, opt_cost, opt_lat) in enumerate(results):
        print(f"  GREEDY[{i}]  ".center(80, '#'))
        evaluate_ser_chain_partitioning(partition, opt_cost, opt_lat, runtime, memory, rate, data, M, L, start, end,
                                        delay)
    return results


def test_chain():
    params = dict(runtime=[20, 40, 50, 20, 70, 40, 50, 60, 40, 10],
                  memory=[3, 3, 2, 1, 2, 1, 2, 1, 2, 3],
                  rate=[1, 1, 2, 2, 1, 3, 1, 2, 1, 3],
                  data=[5, 3, 5, 2, 1, 3, 2, 3, 5, 1],
                  delay=10,
                  M=6,
                  L=800,
                  start=0,
                  end=9)
    print_ser_chain_summary(params['runtime'], params['memory'], params['rate'], params['data'])
    run_test(**params)


def test_partial_ser_chain():
    params = dict(runtime=[20, 40, 50, 20, 70, 40, 50, 60, 40, 10],
                  memory=[3, 3, 2, 1, 2, 1, 2, 1, 2, 3],
                  rate=[1, 1, 2, 2, 1, 3, 1, 2, 1, 3],
                  data=[5, 3, 5, 2, 1, 3, 2, 3, 5, 1],
                  delay=10,
                  M=6,
                  L=math.inf,
                  start=0,
                  end=9)
    print_ser_chain_summary(params['runtime'], params['memory'], params['rate'], params['data'])
    # No restriction
    run_test(**params)
    # Optimal
    params['L'] = 702
    run_test(**params)
    # Stricter restriction
    for l in [600, 540, 535, 525, 465]:
        params['L'] = l
        print(f"  New latency limit({l})  ".center(80, '='))
        run_test(**params)
    # Infeasible due to M
    params['L'] = 455
    run_test(**params)


def test_random_chain(n: int = 10):
    runtime, memory, rate, data = get_random_chain_data(n, (10, 100), (1, 3), (1, 3), (1, 5))
    params = dict(runtime=runtime,
                  memory=memory,
                  rate=rate,
                  data=data,
                  delay=10,
                  M=6,
                  L=700,
                  start=0,
                  end=n - 1)
    print_ser_chain_summary(runtime, memory, rate, data)
    run_test(**params)


if __name__ == '__main__':
    # test_chain()
    # test_random_chain()
    test_partial_ser_chain()



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math
import typing

# Service graph attributes
PLATFORM = 'P'
NAME = 'name'

# Node properties
RUNTIME = 'time'
MEMORY = 'mem'
# Optional, denoting the vCPU core demands in case of multi-threaded functions
CPU = 'cpu'

# Edge properties
RATE = 'rate'
DATA = 'data'
WEIGHT = 'weight'

# Additional attributes for plotting and preprocessing
COLOR = 'color'
LABEL = 'label'

# Separator character for state names, separating nodes in the group (VSEP) and group from assigned flavor (FSEP)
SEP = '|'
ASSIGN = '@'


class Flavor(typing.NamedTuple):
    """Store subtree partitioning attributes for a given subcase."""
    mem: int = math.inf  # Available memory
    ncore: int = 1  # Available relative vCPU cores
    cfactor: float = 1.0  # Relative cost factor

    def __repr__(self):
        # return repr(tuple(self))
        return self.name

    @property
    def name(self) -> str:
        """String representation of the given flavor."""
        return f"F[{self.mem},{self.ncore},{self.cfactor}]"



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
import itertools
import math
import typing

import numpy as np

from slambuc.alg import INFEASIBLE, T_BARRS, MEM, CPU, COST, LAT, BARR


class State(typing.NamedTuple):
    """Store block attributes for a given DP subcase."""
    barr: int = None  # Barrier/heading node of the last block in the given subcase partitioning
    cost: int = math.inf  # Sum cost of the partitioning
    lat: int = math.inf  # Sum latency of the partitioning regarding the limited subchain[start, end]

    def __repr__(self):
        return repr(tuple(self))


def chain_partitioning(runtime: list, memory: list, rate: list, M: int = math.inf, N: int = math.inf,
                       L: int = math.inf, start: int = 0, end: int = None, delay: int = 1,
                       unit: int = 100, ret_dp: bool = False) -> tuple[T_BARRS | list[list[State]], int, int]:
    """
    Calculates minimal-cost partitioning of a chain based on the node properties of *running time*, *memory usage* and
    *invocation rate* with respect to an upper bound **M** on the total memory of blocks and a latency constraint **L**
    defined on the subchain between *start* and *end* nodes.

    Cost calculation relies on the rounding *unit* and number of vCPU cores *N*, whereas platform invocation *delay*
    is used for latency calculations.

    Details in: J. Czentye, I. Pelle and B. Sonkoly, "Cost-optimal Operation of Latency Constrained Serverless
    Applications: From Theory to Practice," NOMS 2023-2023 IEEE/IFIP Network Operations and Management Symposium,
    Miami, FL, USA, 2023, pp. 1-10, doi: 10.1109/NOMS56928.2023.10154412.

    :param runtime: running times in ms
    :param memory:  memory requirements in MB
    :param rate:    avg. rate of function invocations
    :param M:       upper memory bound of the partition blocks (in MB)
    :param N:       upper CPU core bound of the partition blocks
    :param L:       latency limit defined on the critical path in the form of subchain[start -> end] (in ms)
    :param delay:   invocation delay between blocks
    :param start:   head node of the latency-limited subchain
    :param end:     tail node of the latency-limited subchain
    :param unit:    rounding unit for the cost calculation (default: 100 ms)
    :param ret_dp:  return the calculated DP matrix instead of barrier nodes
    :return:        tuple of barrier nodes, sum cost of the partitioning, and the calculated latency on the subchain
    """
    n = len(runtime)
    end = end if end is not None else n - 1

    @functools.lru_cache(maxsize=n - 1)
    def block_memory(_b: int, _w: int) -> int:
        """Calculate memory of block[b, w]"""
        return sum(memory[_b: _w + 1])

    @functools.lru_cache(maxsize=n - 1)
    def block_cpu(_b: int, _w: int) -> int:
        """Calculate memory of block[b, w]"""
        r_max = itertools.chain((1,), enumerate(itertools.accumulate(reversed(rate[_b: _w + 1]), max)))
        return functools.reduce(lambda pre, max_i: max(pre, math.ceil(max_i[1] / rate[_w - max_i[0]])), r_max)

    @functools.lru_cache(maxsize=n - 1)
    def block_cost(_b: int, _w: int) -> int:
        """Calculate running time of block[b, w]"""
        return rate[_b] * (math.ceil(sum(runtime[_b: _w + 1]) / unit) * unit)

    @functools.lru_cache(maxsize=n - 1)
    def block_latency(_b: int, _w: int) -> int:
        """Calculate relevant latency for block[b, w]"""
        # Do not consider latency if no intersection
        if end < _b or _w < start:
            return 0
        blk_lat = sum(runtime[max(_b, start): min(_w, end) + 1])
        # Ignore delay if latency path starts within the subchain
        return delay + blk_lat if start < _b else blk_lat

    # Check lower bound for latency limit
    if L < sum(runtime[start: end + 1]):
        return INFEASIBLE
    # Check if memory constraint allows feasible solutions for the given latency constraint
    k_min = max(math.ceil(sum(memory[start: end + 1]) / M),
                sum(1 for i, j in itertools.pairwise(rate) if math.ceil(j / i) > N))
    k_max = math.floor(min((L - sum(runtime[start: end + 1])) / delay + 1, n))
    if k_max < k_min:
        return INFEASIBLE
    # Check single node partitioning
    if len(runtime) == 1:
        return [0], block_cost(0, 0), block_latency(0, 0)
    # Initialize left triangular part of DP matrix -> DP[i][j][COST, LAT, BARR]
    DP = [[State() for _ in range(i + 1)] for i in range(n)]
    # Initialize default values for grouping first w nodes into one group
    for w in range(0, n):
        if block_memory(0, w) > M or block_cpu(0, w) > N:
            break
        DP[w][0] = State(0, block_cost(0, w), block_latency(0, w))
    # Calculate Dynamic Programming matrix
    for w in range(1, n):
        for k in range(1, w + 1):
            for b in reversed(range(k, w + 1)):
                # As k decreases, bigger blocks [k, w] will continue violating the memory constraint
                if block_memory(b, w) > M or block_cpu(b, w) > N:
                    break
                if (lat := DP[b - 1][k - 1].lat + block_latency(b, w)) <= L:
                    # Store and overwrite subcases with equal costs (<=) to consider larger blocks for lower latency
                    if (cost := DP[b - 1][k - 1].cost + block_cost(b, w)) <= DP[w][k].cost:
                        DP[w][k] = State(b, cost, lat)
            # If first w node cannot be partitioned into k blocks due to L then it cannot be partitioned into k+1
            if DP[w][k - 1].lat < DP[w][k].lat == math.inf:
                break
    # Index of optimal cost partition, the fist one if multiple min values exist
    k_opt = min(range(n), key=lambda x: DP[-1][x].cost)
    _, opt_cost, opt_lat = DP[-1][k_opt]
    if opt_cost < math.inf:
        return DP if ret_dp else extract_barr(DP, k_opt), opt_cost, opt_lat
    else:
        return INFEASIBLE


def extract_barr(DP: list[list[State]], k: int) -> T_BARRS:
    """
    Extract barrier nodes form DP matrix by iteratively backtracking the minimal cost subcases started from *k*.

    :param DP:  DP matrix containing subcase *States*
    :param k:   number of optimal cuts
    :return:    list of barrier nodes
    """
    barr = []
    w = len(DP) - 1
    for k in reversed(range(0, k + 1)):
        # The cached b value marks the barrier node of the k. block and refers the subcase => C[b-1,k-1] + c[b,w]
        b = DP[w][k].barr
        w = b - 1
        barr.append(b)
    barr.reverse()
    return barr


########################################################################################################################


def vec_chain_partitioning(runtime: list, memory: list, rate: list, M: int = np.inf, N: int = np.inf, L: int = np.inf,
                           start: int = 0, end: int = None, delay: int = 1, unit: int = 100,
                           ret_dp: bool = False) -> tuple[T_BARRS | np.ndarray, int, int]:
    """
    Calculates minimal-cost partitioning of a chain based on the node properties of *runtime*, *memory* and *rate* with
    respect to an upper bound **M** on the total memory of blocks and a latency constraint **L** defined on the subchain
    between *start* and *end* nodes leveraging vectorized operations.

    Cost calculation relies on the rounding *unit* and number of vCPU cores *N*, whereas platform invocation *delay*
    is used for latency calculations.

    Details in: J. Czentye, I. Pelle and B. Sonkoly, "Cost-optimal Operation of Latency Constrained Serverless
    Applications: From Theory to Practice," NOMS 2023-2023 IEEE/IFIP Network Operations and Management Symposium,
    Miami, FL, USA, 2023, pp. 1-10, doi: 10.1109/NOMS56928.2023.10154412.

    :param runtime: running times in ms
    :param memory:  memory requirements in MB
    :param rate:    avg. rate of function invocations
    :param M:       upper memory bound of the partition blocks (in MB)
    :param N:       upper CPU core bound of the partition blocks
    :param L:       latency limit defined on the critical path in the form of subchain[start -> end] (in ms)
    :param delay:   invocation delay between blocks
    :param start:   head node of the latency-limited subchain
    :param end:     tail node of the latency-limited subchain
    :param unit:    rounding unit for the cost calculation (default: 100 ms)
    :param ret_dp:  return the calculated DP matrix instead of the barrier nodes
    :return:        tuple of barrier nodes, sum cost of the partitioning, and the calculated latency on the subchain
    """
    n = len(runtime)
    end = end if end is not None else n - 1

    def block_memory(_v: int, cumsum: list = (0, 0, 0), from_left=False) -> int:
        """Calculate memory of a block from the cumulative sum of block[v+1, w] or block[b, v-1]"""
        cumsum[MEM] += memory[_v]
        return cumsum[MEM]

    def block_cpu(_v: int, cumsum: list = (0, 0, 0, (0, 1)), from_left=False) -> int:
        """Calculate CPU need of a block from the cached cumulative sum of block[v+1, w]"""
        cumsum[CPU][0] = max(cumsum[CPU][0], rate[_v])
        if from_left:
            return cumsum[CPU][0]
        cumsum[CPU][1] = max(cumsum[CPU][1], math.ceil(cumsum[CPU][0] / rate[_v]))
        return cumsum[CPU][1]

    def block_cost(_v: int, cumsum: list = (0, 0, 0), from_left: bool = False) -> int:
        """Calculate running time of block[b, w] from the cumulative sum of block[v+1, w] or block[0, v-1]"""
        cumsum[COST] += runtime[_v]
        return rate[0 if from_left else _v] * (math.ceil(cumsum[COST] / unit) * unit)

    def block_latency(_b: int, _w: int, cumsum: list = (0, 0, 0), from_left: bool = False) -> int:
        """Calculate relevant latency for block[b, w] from the cumulative sum of block[b+1, w] or block[0, w-1]"""
        # Do not consider latency if no intersection
        if end < _b or _w < start:
            return 0
        if from_left:
            # No need to add next node if it is outside of intersection
            if end < _w:
                return cumsum[LAT]
            cumsum[LAT] += runtime[_w]
        else:
            # No need to add next node if it is outside of intersection
            if b < start:
                return cumsum[LAT]
            cumsum[LAT] += runtime[_b]
        # Ignore delay if latency path starts within the subchain
        return delay + cumsum[LAT] if start < _b else cumsum[LAT]

    # Check lower bound for latency limit
    if L < sum(runtime[start: end + 1]):
        return INFEASIBLE
    # Check if memory constraint allows feasible solutions for the given latency constraint
    k_min = max(math.ceil(sum(memory[start: end + 1]) / M),
                sum(1 for i, j in itertools.pairwise(rate) if math.ceil(j / i) > N))
    k_max = math.floor(min((L - sum(runtime[start: end + 1])) / delay + 1, n))
    if k_max < k_min:
        return INFEASIBLE
    # Check single node partitioning
    if len(runtime) == 1:
        return [0], block_cost(0), block_latency(0, 0)
    # Initialize DP matrix -> DP[i][j][BARR, COST, LAT]
    DP = np.full((n, n, 3), np.inf)
    # Define cache for cumulative block attribute calculations: [MEM, COST, LAT]
    __cache = [0, 0, 0, [0, 0]]
    # Initialize default values for grouping first w nodes into one group
    for w in range(0, n):
        if block_memory(w, __cache, from_left=True) > M or block_cpu(w, __cache, from_left=True) > N:
            break
        DP[w, 0] = np.array((0, block_cost(w, __cache, from_left=True), block_latency(0, w, __cache, from_left=True)))
    # Calculate Dynamic Programming matrix
    for w in range(1, n):
        # Cache for the cumulative sum based calculation of expanding block's memory, runtime and latency values
        __cache = [0, 0, 0, [0, 0]]
        for b in reversed(range(1, w + 1)):
            # As k decreases, bigger blocks [k, w] will continue violating the memory constraint
            if block_memory(b, __cache) > M or block_cpu(b, __cache) > N:
                break
            subcases = DP[b - 1, :b] + np.array((0, block_cost(b, __cache), block_latency(b, w, __cache)))
            subcases[:, BARR] = b
            # Store and overwrite subcases with equal costs (<=) to consider larger blocks for lower latency
            feasible_idx = np.flatnonzero((subcases[:, LAT] <= L) & (subcases[:, COST] <= DP[w, 1:b + 1, COST]))
            DP[w, feasible_idx + 1] = subcases[feasible_idx]
    # Index of optimal cost partition, the fist one if multiple min values exist
    k_opt = np.argmin(DP[n - 1, :, COST])
    _, opt_cost, opt_lat = DP[n - 1, k_opt]
    if opt_cost < np.inf:
        return DP if ret_dp else extract_vec_barr(DP, k_opt), opt_cost, opt_lat
    else:
        return INFEASIBLE


def extract_vec_barr(DP: np.ndarray, k: int) -> T_BARRS:
    """
    Extract barrier nodes from vectorized DP matrix by iteratively backtracking the minimal cost subcases from *k*.

    :param DP:  DP matrix containing subcase *States*
    :param k:   number of optimal cuts
    :return:    list of barrier nodes
    """
    barr = []
    B = DP[..., BARR]
    w = len(B) - 1
    for k in reversed(range(0, k + 1)):
        # The cached b value marks the barrier node of the k. block and refers the subcase => C[b-1,k-1] + c[b,w]
        b = int(B[w, k])
        w = b - 1
        barr.append(b)
    barr.reverse()
    return barr



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pathlib

import networkx as nx

from slambuc.alg.ext.greedy import (min_weight_greedy_partitioning, get_feasible_cpath_split,
                                    min_weight_partition_heuristic, min_lat_partition_heuristic, get_min_cpath_split)
from slambuc.alg.util import ibacktrack_chain
from slambuc.misc.generator import get_random_tree
from slambuc.misc.util import evaluate_par_tree_partitioning


def test_min_weight_bounded_partitioning(root: int = 1, M: int = 11, L: int = 450, cp_end: int = 10, N: int = 1,
                                         delay: int = 10):
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree.gml", destringizer=int)
    # tree = get_random_tree(10)
    partition, sum_cost, sum_lat = min_weight_greedy_partitioning(tree, root, M, N)
    evaluate_par_tree_partitioning(tree, partition, sum_cost, sum_lat, root, cp_end, M, L, N, delay=delay)


def test_min_weight_partition_heuristic(root: int = 1, M: int = 15, L: int = 475, cp_end: int = 10, N: int = 1,
                                        delay: int = 10):
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree.gml", destringizer=int)
    # tree = get_random_tree(10)
    print("Cpath:", list(reversed(list(ibacktrack_chain(tree, root, cp_end)))))
    split = get_feasible_cpath_split(tree, root, cp_end, M, L, N, delay)
    print("Calculated split:", split)
    partition, sum_cost, sum_lat = min_weight_partition_heuristic(tree, root, M, L, N, cp_end, delay)
    evaluate_par_tree_partitioning(tree, partition, sum_cost, sum_lat, root, cp_end, M, L, N, delay=delay)


def test_min_lat_partition_heuristic(root: int = 1, M: int = 15, L: int = 475, cp_end: int = 10, N: int = 1,
                                     delay: int = 10):
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree.gml", destringizer=int)
    # tree = get_random_tree(10)
    print("Cpath:", list(reversed(list(ibacktrack_chain(tree, root, cp_end)))))
    split = get_min_cpath_split(tree, root, cp_end, M, L, N, delay)
    print("Calculated split:", split)
    partition, sum_cost, sum_lat = min_lat_partition_heuristic(tree, root, M, L, N, cp_end, delay)
    evaluate_par_tree_partitioning(tree, partition, sum_cost, sum_lat, root, cp_end, M, L, N, delay=delay)


def compare_heuristics(root: int = 1, M: int = 15, L: int = 475, cp_end: int = 10, N: int = 1, delay: int = 10):
    tree = get_random_tree(10)
    print("Cpath:", list(reversed(list(ibacktrack_chain(tree, root, cp_end)))))
    print("Min weight partitioning")
    partition, sum_cost, sum_lat = min_weight_partition_heuristic(tree, root, M, L, N, cp_end, delay)
    evaluate_par_tree_partitioning(tree, partition, sum_cost, sum_lat, root, cp_end, M, L, N, delay=delay)
    #
    print("Min lat partitioning")
    partition, sum_cost, sum_lat = min_lat_partition_heuristic(tree, root, M, L, N, cp_end, delay)
    evaluate_par_tree_partitioning(tree, partition, sum_cost, sum_lat, root, cp_end, M, L, N, delay=delay)


if __name__ == '__main__':
    # test_min_weight_bounded_partitioning()
    # test_min_weight_partition_heuristic()
    # test_min_lat_partition_heuristic()
    compare_heuristics()



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math
import pathlib
import pprint
import time

import networkx as nx
import pulp

from slambuc.alg import LP_LAT
from slambuc.alg.service import NAME
from slambuc.alg.tree.par.ilp import (build_greedy_par_tree_mtx_model, build_par_tree_mtx_model,
                                      tree_par_mtx_partitioning)
from slambuc.alg.tree.ser.ilp import extract_subtrees_from_xmatrix
from slambuc.alg.util import ibacktrack_chain
from slambuc.misc.generator import get_random_tree
from slambuc.misc.util import (print_lp_desc, print_var_matrix, evaluate_par_tree_partitioning,
                               print_pulp_matrix_values, convert_var_dict, print_cost_coeffs, print_lat_coeffs)


def test_par_mtx_model_creation(tree_file: str = pathlib.Path(__file__).parent / "data/graph_test_tree_par.gml",
                                save_file: bool = False):
    tree = nx.read_gml(tree_file, destringizer=int)
    tree.graph[NAME] += "-par_ilp_mtx"
    cpath = set(ibacktrack_chain(tree, 1, 10))
    params = dict(tree=tree,
                  root=1,
                  cpath=cpath,
                  M=6,
                  L=430,
                  N=2,
                  delay=10)
    print("  Test input  ".center(80, '='))
    pprint.pprint(params)
    print('=' * 80)
    _s = time.perf_counter()
    model_greedy, X_greedy = build_greedy_par_tree_mtx_model(**params)
    _d = time.perf_counter() - _s
    print(f"Greedy Model building time: {_d * 1000} ms")
    _s = time.perf_counter()
    model, X = build_par_tree_mtx_model(**params)
    _d = time.perf_counter() - _s
    print(f"Direct Model building time: {_d * 1000} ms")
    X = convert_var_dict(X)
    X_greedy = convert_var_dict(X_greedy)
    print("  Decision variables  ".center(80, '='))
    print("Greedy:")
    print_var_matrix(X_greedy)
    print("Direct:")
    print_var_matrix(X)
    print("  Cost Coefficients  ".center(80, '='))
    print("Greedy:")
    print_cost_coeffs(model_greedy, X_greedy)
    print("Direct:")
    print_cost_coeffs(model, X)
    print("  Latency Coefficients  ".center(80, '='))
    print("Greedy:")
    print_lat_coeffs(model_greedy, X_greedy)
    print("Direct:")
    print_lat_coeffs(model, X)
    print("  Generated LP model  ".center(80, '='))
    print_lp_desc(model_greedy)
    print_lp_desc(model)
    if save_file:
        model.writeLP("tree_par_mtx_model.lp")


def test_par_mtx_model_solution():
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree_par.gml", destringizer=int)
    tree.graph[NAME] += "-par_ilp_mtx"
    cpath = set(ibacktrack_chain(tree, 1, 10))
    params = dict(tree=tree,
                  root=1,
                  cpath=cpath,
                  M=6,
                  L=430,
                  N=2,
                  delay=10)
    print("  Run CBC solver  ".center(80, '='))
    #
    print(" Greedy model ".center(80, '='))
    model, X = build_greedy_par_tree_mtx_model(**params)
    status = model.solve(solver=pulp.PULP_CBC_CMD(mip=True, warmStart=False, msg=True))
    print(f"Partitioning status: {status} / {pulp.LpStatus[status]}")
    print(f"Cost/Lat: {pulp.value(model.objective)}, {pulp.value(model.constraints[LP_LAT])}")
    print("Solution:")
    print_pulp_matrix_values(convert_var_dict(X))
    #
    print(" Direct model ".center(80, '='))
    model, X = build_par_tree_mtx_model(**params)
    solver = pulp.PULP_CBC_CMD(mip=True, warmStart=False, msg=True)
    status = model.solve(solver=solver)
    print(f"Partitioning status: {status} / {pulp.LpStatus[status]}")
    print(f"Cost/Lat: {pulp.value(model.objective)}, {pulp.value(model.constraints[LP_LAT])}")
    #
    print("Solution:")
    print_pulp_matrix_values(convert_var_dict(X))
    print(f"Partitioning status: {status} / {pulp.LpStatus[status]}")
    _s = time.perf_counter()
    ext_partition = extract_subtrees_from_xmatrix(X)
    _d = time.perf_counter() - _s
    print(f"Extract:  {_d * 1000} ms")
    print(f"Partitioning: {ext_partition = }")
    print(f"{model.solutionTime    = } s")
    print(f"{model.solutionCpuTime = }")


def evaluate_ilp_par_mtx_model():
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree_par.gml", destringizer=int)
    tree.graph[NAME] += "-par_ilp_mtx"
    params = dict(tree=tree,
                  root=1,
                  cp_end=10,
                  M=6,
                  L=430,
                  N=2,
                  delay=10)
    print("  CBC solver  ".center(80, '='))
    partition, opt_cost, opt_lat = tree_par_mtx_partitioning(**params,
                                                             solver=pulp.PULP_CBC_CMD(mip=True, warmStart=False))
    print(f"Partitioning: {partition}, {opt_cost = }, {opt_lat = }")
    evaluate_par_tree_partitioning(partition=partition, opt_cost=opt_cost, opt_lat=opt_lat, **params)
    print("  CPLEX solver  ".center(80, '='))
    partition, opt_cost, opt_lat = tree_par_mtx_partitioning(**params, solver=pulp.CPLEX_PY(mip=True, warmStart=False))
    print(f"Partitioning: {partition}, {opt_cost = }, {opt_lat = }")
    evaluate_par_tree_partitioning(partition=partition, opt_cost=opt_cost, opt_lat=opt_lat, **params)


def evaluate_ilp_par_subchain_mtx_model():
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree_par.gml", destringizer=int)
    tree.graph[NAME] += "-par_ilp_mtx"
    params = dict(tree=tree,
                  root=1,
                  cp_end=10,
                  M=6,
                  L=430,
                  N=2,
                  delay=10)
    print("  CBC solver  ".center(80, '='))
    partition, opt_cost, opt_lat = tree_par_mtx_partitioning(**params, subchains=True,
                                                             solver=pulp.PULP_CBC_CMD(mip=True, warmStart=False))
    print(f"Partitioning: {partition}, {opt_cost = }, {opt_lat = }")
    evaluate_par_tree_partitioning(partition=partition, opt_cost=opt_cost, opt_lat=opt_lat, **params)
    print("  CPLEX solver  ".center(80, '='))
    partition, opt_cost, opt_lat = tree_par_mtx_partitioning(**params, subchains=True,
                                                             solver=pulp.CPLEX_PY(mip=True, warmStart=False))
    print(f"Partitioning: {partition}, {opt_cost = }, {opt_lat = }")
    evaluate_par_tree_partitioning(partition=partition, opt_cost=opt_cost, opt_lat=opt_lat, **params)


########################################################################################################################

def run_test(tree: nx.DiGraph, root: int, cp_end: int, M: int, L: int, N: int, delay: int):
    partition, opt_cost, opt_lat = tree_par_mtx_partitioning(tree, root, M, L, N, cp_end, delay,
                                                             solver=pulp.PULP_CBC_CMD(mip=True, warmStart=False,
                                                                                      msg=False))
    evaluate_par_tree_partitioning(tree, partition, opt_cost, opt_lat, root, cp_end, M, L, N, delay)
    return partition, opt_cost, opt_lat


def test_par_tree():
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree_par.gml", destringizer=int)
    tree.graph[NAME] += "-par_ilp_mtx"
    params = dict(tree=tree,
                  root=1,
                  cp_end=10,
                  M=6,
                  # L = math.inf
                  L=430,
                  N=2,
                  delay=10)
    run_test(**params)


def test_random_par_tree(n: int = 10):
    tree = get_random_tree(n)
    tree.graph[NAME] += "-par_ilp_mtx"
    params = dict(tree=tree,
                  root=1,
                  cp_end=n,
                  M=6,
                  L=math.inf,
                  N=2,
                  delay=10)
    run_test(**params)


if __name__ == '__main__':
    # test_par_mtx_model_creation()
    # test_par_mtx_model_solution()
    evaluate_ilp_par_mtx_model()
    # evaluate_ilp_par_subchain_mtx_model()
    # test_par_mtx_model_creation(tree_file="failed_random_tree_1675189959.6335204-ser_partition_L358_M6.gml")
    # test_par_tree()
    # test_random_par_tree()



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math
import pathlib
import pprint
import time

import networkx as nx
import pulp

from slambuc.alg import LP_LAT
from slambuc.alg.service import NAME
from slambuc.alg.service.common import Flavor
from slambuc.alg.tree.layout.ilp import (build_gen_tree_cfg_model, recreate_st_from_gen_xdict, build_gen_tree_mtx_model,
                                         extract_st_from_gen_xmatrix, tree_gen_hybrid_partitioning,
                                         tree_gen_mtx_partitioning, all_gen_tree_mtx_partitioning)
from slambuc.misc.generator import get_random_tree
from slambuc.misc.util import (print_lp_desc, convert_var_dict, print_var_matrix, print_cost_coeffs, print_lat_coeffs,
                               print_pulp_matrix_values, evaluate_gen_tree_partitioning)


def test_gen_cfg_model_creation(tree_file: str = pathlib.Path(__file__).parent / "data/graph_test_tree_par.gml",
                                save_file: bool = False):
    tree = nx.read_gml(tree_file, destringizer=int)
    tree.graph[NAME] += "-gen_ilp_cfg"
    params = dict(tree=tree,
                  root=1,
                  cp_end=10,
                  flavors=[Flavor(6, 1, 1),
                           Flavor(9, 3, 1.1)],  # Higher memory with more core, but 10% more expensive
                  L=430,
                  delay=10)
    print("  Test input  ".center(80, '='))
    pprint.pprint(params)
    print('=' * 80)
    _s = time.perf_counter()
    model, X = build_gen_tree_cfg_model(**params)
    _d = time.perf_counter() - _s
    print(f"General CFG model building time: {_d * 1000} ms")
    print("Distribution of configurations:")
    for flavor in X:
        print(f"#### Flavor: {flavor}")
        for i in X[flavor]:
            print(f"{i} -> {len(X[flavor][i])}")
    print("  Generated LP model  ".center(80, '='))
    print_lp_desc(model)
    if save_file:
        model.writeLP("tree_gen_model.lp")


def test_gen_cfg_model_solution():
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree_par.gml", destringizer=int)
    tree.graph[NAME] += "-gen_ilp_cfg"
    params = dict(tree=tree,
                  root=1,
                  cp_end=10,
                  flavors=[Flavor(6, 1, 1),
                           Flavor(9, 3, 1.1)],  # Higher memory with more core, but 10% more expensive
                  L=430,
                  delay=10)
    print("  Run CBC solver  ".center(80, '='))
    #
    model, X = build_gen_tree_cfg_model(**params)
    solver = pulp.PULP_CBC_CMD(mip=True, warmStart=False, msg=True)
    status = model.solve(solver=solver)
    print(f"Partitioning status: {status} / {pulp.LpStatus[status]}")
    print(f"Cost/Lat: {pulp.value(model.objective)}, {pulp.value(model.constraints[LP_LAT])}")
    print(f"Partitioning status: {status} / {pulp.LpStatus[status]}")
    _s = time.perf_counter()
    ext_partition = recreate_st_from_gen_xdict(tree, X)
    _d = time.perf_counter() - _s
    print(f"Extract:  {_d * 1000} ms")
    print(f"Partitioning: {ext_partition = }")
    print(f"{model.solutionTime    = } s")
    print(f"{model.solutionCpuTime = }")


def test_gen_mtx_model_creation(tree_file: str = pathlib.Path(__file__).parent / "data/graph_test_tree_par.gml",
                                save_file: bool = False):
    tree = nx.read_gml(tree_file, destringizer=int)
    tree.graph[NAME] += "-gen_ilp_mtx"
    params = dict(tree=tree,
                  root=1,
                  cp_end=10,
                  flavors=[Flavor(6, 1, 1),
                           Flavor(9, 3, 1.1)],  # Higher memory with more core, but 10% more expensive
                  L=430,
                  delay=10)
    print("  Test input  ".center(80, '='))
    pprint.pprint(params)
    print('=' * 80)
    _s = time.perf_counter()
    model, X = build_gen_tree_mtx_model(**params)
    _d = time.perf_counter() - _s
    print(f"General MTX model building time: {_d * 1000} ms")
    #
    X_f = [convert_var_dict(X[f]) for f in X]
    print("  Decision variables  ".center(80, '='))
    for f, X in zip(params['flavors'], X_f):
        print(f"Flavor: {f}")
        print_var_matrix(X)
    print("  Cost Coefficients  ".center(80, '='))
    for f, X in zip(params['flavors'], X_f):
        print(f"Flavor: {f}")
        print_cost_coeffs(model, X)
    print("  Latency Coefficients  ".center(80, '='))
    for f, X in zip(params['flavors'], X_f):
        print(f"Flavor: {f}")
        print_lat_coeffs(model, X)
    print("  Generated LP model  ".center(80, '='))
    print_lp_desc(model)
    if save_file:
        model.writeLP("tree_gen_mtx_model.lp")


def test_gen_mtx_model_solution():
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree_par.gml", destringizer=int)
    tree.graph[NAME] += "-par_ilp_mtx"
    params = dict(tree=tree,
                  root=1,
                  cp_end=10,
                  flavors=[Flavor(6, 1, 1),
                           Flavor(9, 3, 1.1)],  # Higher memory with more core, but 10% more expensive
                  L=430,
                  delay=10)
    print("  Run CBC solver  ".center(80, '='))
    #
    print(" General MTX model ".center(80, '='))
    model, X = build_gen_tree_mtx_model(**params)
    solver = pulp.PULP_CBC_CMD(mip=True, warmStart=False, msg=True)
    status = model.solve(solver=solver)
    print(f"Partitioning status: {status} / {pulp.LpStatus[status]}")
    print(f"Cost/Lat: {pulp.value(model.objective)}, {pulp.value(model.constraints[LP_LAT])}")
    print("Solution:")
    for f in X:
        print(f"Flavor: {f}")
        print_pulp_matrix_values(convert_var_dict(X[f]))
    #
    _s = time.perf_counter()
    ext_partition = extract_st_from_gen_xmatrix(X)
    _d = time.perf_counter() - _s
    print(f"Extract:  {_d * 1000} ms")
    print(f"Partitioning: {ext_partition = }")
    print(f"{model.solutionTime    = } s")
    print(f"{model.solutionCpuTime = }")


def test_gen_mtx_subchain_model_solution():
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree_par.gml", destringizer=int)
    tree.graph[NAME] += "-par_ilp_mtx"
    params = dict(tree=tree,
                  root=1,
                  cp_end=10,
                  flavors=[Flavor(6, 1, 1),
                           Flavor(9, 3, 1.1)],  # Higher memory with more core, but 10% more expensive
                  L=430,
                  subchains=True,
                  delay=10)
    print("  Run CBC solver  ".center(80, '='))
    #
    print(" General MTX model ".center(80, '='))
    model, X = build_gen_tree_mtx_model(**params)
    solver = pulp.PULP_CBC_CMD(mip=True, warmStart=False, msg=True)
    status = model.solve(solver=solver)
    print(f"Partitioning status: {status} / {pulp.LpStatus[status]}")
    print(f"Cost/Lat: {pulp.value(model.objective)}, {pulp.value(model.constraints[LP_LAT])}")
    print("Solution:")
    for f in X:
        print(f"Flavor: {f}")
        print_pulp_matrix_values(convert_var_dict(X[f]))
    #
    _s = time.perf_counter()
    ext_partition = extract_st_from_gen_xmatrix(X)
    _d = time.perf_counter() - _s
    print(f"Extract:  {_d * 1000} ms")
    print(f"Partitioning: {ext_partition = }")
    print(f"{model.solutionTime    = } s")
    print(f"{model.solutionCpuTime = }")
    evaluate_gen_tree_partitioning(tree, ext_partition, 0, 0, 0, [[0, 0, 0]], 0, 0, 0)


########################################################################################################################

def run_test(tree: nx.DiGraph, root: int, flavors: list[Flavor], cp_end: int, L: int, delay: int):
    partition, opt_cost, opt_lat = tree_gen_hybrid_partitioning(tree, root, flavors, L=L, cp_end=cp_end, delay=delay,
                                                                solver=pulp.PULP_CBC_CMD(mip=True, warmStart=False,
                                                                                         msg=False))
    evaluate_gen_tree_partitioning(tree, partition, opt_cost, opt_lat, root, flavors, cp_end, L, delay)
    partition, opt_cost, opt_lat = tree_gen_mtx_partitioning(tree, root, flavors, L=L, cp_end=cp_end, delay=delay,
                                                             solver=pulp.PULP_CBC_CMD(mip=True, warmStart=False,
                                                                                      msg=False))
    evaluate_gen_tree_partitioning(tree, partition, opt_cost, opt_lat, root, flavors, cp_end, L, delay)
    return partition, opt_cost, opt_lat


def test_gen_tree():
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree_par.gml", destringizer=int)
    tree.graph[NAME] += "-gen_ilp"
    params = dict(tree=tree,
                  root=1,
                  cp_end=10,
                  flavors=[Flavor(6, 1, 1),
                           Flavor(9, 3, 1.1)],  # Higher memory with more core, but 10% more expensive
                  L=430,
                  delay=10)
    run_test(**params)


def test_random_gen_tree(n: int = 10):
    tree = get_random_tree(n)
    tree.graph[NAME] += "-gen_ilp"
    params = dict(tree=tree,
                  root=1,
                  cp_end=10,
                  flavors=[Flavor(6, 1, 1),
                           Flavor(9, 3, 1.1)],  # Higher memory with more core, but 10% more expensive
                  L=math.inf,
                  delay=10)
    run_test(**params)


def test_all_gen_tree():
    tree = nx.read_gml(pathlib.Path(__file__).parent / "data/graph_test_tree_par.gml", destringizer=int)
    tree.graph[NAME] += "-gen_ilp"
    params = dict(tree=tree,
                  root=1,
                  cp_end=10,
                  flavors=[Flavor(6, 1, 1),
                           Flavor(9, 3, 1.1)],  # Higher memory with more core, but 10% more expensive
                  L=430,
                  delay=10)
    results = all_gen_tree_mtx_partitioning(solver=pulp.PULP_CBC_CMD(mip=True, warmStart=False, msg=False), **params)
    print("Results:")
    for partition, opt_cost, opt_lat in results:
        evaluate_gen_tree_partitioning(partition=partition, opt_cost=opt_cost, opt_lat=opt_lat, **params)


if __name__ == '__main__':
    # test_gen_cfg_model_creation()
    # test_gen_cfg_model_solution()
    # test_gen_mtx_model_creation()
    # test_gen_mtx_model_solution()
    test_gen_mtx_subchain_model_solution()
    #
    # test_gen_tree()
    # test_random_gen_tree()
    #
    # test_all_gen_tree()



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import collections
import itertools
import math
import typing

import networkx as nx

from slambuc.alg import COST, INFEASIBLE, T_BRESULTS
from slambuc.alg.chain.dp.mtx import chain_partitioning, extract_barr
from slambuc.alg.service import *
from slambuc.alg.service.common import LABEL
from slambuc.alg.util import ipostorder_dfs, isubchains, ibacktrack_chain, leaf_label_nodes, recreate_subchain_blocks


class TPart(typing.NamedTuple):
    """Store subtree attributes for a given meta subcase."""
    barr: set = set()  # Barrier/heading nodes of the given subtree partitioning
    cost: int = math.inf  # Sum cost of the partitioning

    def __repr__(self):
        return repr(tuple(self))


def meta_tree_partitioning(tree: nx.DiGraph, root: int = 1, M: int = math.inf, N: int = math.inf, L: int = math.inf,
                           cp_end: int = None, delay: int = 1, unit: int = 100, only_barr: bool = False,
                           partition=chain_partitioning, barriers=extract_barr) -> T_BRESULTS:
    """
    Calculates minimal-cost partitioning of a service graph(tree) with respect to an upper bound **M** on the total
    memory of blocks and a latency constraint **L** defined on the subchain between *root* and *cp_end* nodes using
    the *partition* function to partition subchains independently.

    Cost calculation relies on the rounding *unit* and number of vCPU cores *N*, whereas platform invocation *delay*
    is used for latency calculations.

    It gives optimal result only in case the cost function regarding the chain attributes is sub-additive,
    that is k_opt = k_min is guaranteed for each case.

    Instead of full partitioning it only returns the list of barrier nodes.

    Details in: J. Czentye, I. Pelle and B. Sonkoly, "Cost-optimal Operation of Latency Constrained Serverless
    Applications: From Theory to Practice," NOMS 2023-2023 IEEE/IFIP Network Operations and Management Symposium,
    Miami, FL, USA, 2023, pp. 1-10, doi: 10.1109/NOMS56928.2023.10154412.

    :param tree:        service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param root:        root node of the graph
    :param M:           upper memory bound of the partition blocks (in MB)
    :param N:           upper CPU core bound of the partition blocks
    :param L:           latency limit defined on the critical path (in ms)
    :param cp_end:      tail node of the critical path in the form of subchain[root -> cp_end]
    :param delay:       invocation delay between blocks
    :param unit:        rounding unit for the cost calculation (default: 100 ms)
    :param only_barr:   return the subtree roots (barrier nodes) instead of full node partitioning
    :param partition:   function that partitions chain into blocks wrt. M and L
    :param barriers:    function that extracts barrier nodes from _partition_'s returned data structure
    :return:            tuple of barrier nodes, sum cost of the partitioning, and optimal number of cuts
    """
    tree = leaf_label_nodes(tree)
    cpath = set(ibacktrack_chain(tree, root, cp_end))
    # c_max is the number of cuts allowed by L or at most the number of edges on cpath
    c_max = math.floor(min((L - sum(tree.nodes[v][RUNTIME] for v in cpath)) / delay, len(cpath) - 1))
    # Check lower bound for latency limit
    if c_max < 0:
        return [], None, c_max
    DP = collections.defaultdict(lambda: [TPart() for _ in range(c_max + 1)])
    for pred, n in ipostorder_dfs(tree, root):
        # Only branched nodes are referred in the subcases
        if len(tree.succ[pred]) <= 1 and pred is not PLATFORM:
            continue
        # Subcases of leaves can be precalculated
        if n in tree.nodes[root][LABEL]:
            # For single nodes feasible solution exists by definition
            _, opt_cost, _ = partition([tree.nodes[n][RUNTIME]], [tree.nodes[n][MEMORY]], [tree[pred][n][RATE]], M, N,
                                       delay=delay, unit=unit)
            DP[n] = [TPart({n}, opt_cost)] * (c_max if n == cp_end else 1)
            continue
        for (head_part, tail_part), branches in isubchains(tree, n, cp_end):
            subchain = head_part + tail_part
            runtime, memory, rate = zip(*[(tree.nodes[v][RUNTIME], tree.nodes[v][MEMORY], tree[u][v][RATE])
                                          for u, v in itertools.pairwise([pred, *subchain])])
            sum_m_cost = sum(DP[m][0].cost for m in branches if m not in cpath)
            sum_m_barr = set().union(*(DP[m][0].barr for m in branches if m not in cpath))
            # Subchain has no intersection with cpath -> no need to track cuts
            if n not in cpath:
                # Without L, there should exist feasible solution wrt. M
                barr, opt_cost, _ = partition(runtime, memory, rate, M, N, delay=delay, unit=unit)
                if (sum_cost := opt_cost + sum_m_cost) < DP[n][0].cost:
                    DP[n][0] = TPart({subchain[b] for b in barr} | sum_m_barr, sum_cost)
            # Subchain is the tail part of cpath -> all inner edges are allowed to be merged
            elif subchain[-1] == cp_end:
                # If subchain is the cpath -> partitioning can be calculated directly applying L
                if subchain[0] == root:
                    barr, opt_cost, opt_lat = partition(runtime, memory, rate, M, N, L, delay=delay, unit=unit)
                    if not barr:
                        return barr, opt_cost, opt_lat
                    opt_barr_cost = TPart({subchain[b] for b in barr} | sum_m_barr, sum_cost := opt_cost + sum_m_cost)
                    for c in range(len(barr) - 1, c_max + 1):
                        if sum_cost < DP[n][c].cost:
                            DP[n][c] = opt_barr_cost
                else:
                    # Without L, there should exist feasible solution wrt. M
                    CDP, *_ = partition(runtime, memory, rate, M, N, delay=delay, unit=unit, ret_dp=True)
                    c_best, part_best = 0, None
                    for c in range(c_max + 1):
                        if c < len(subchain):
                            # No need to track infeasible solutions
                            if CDP[-1][c][COST] == math.inf:
                                continue
                            # If c-1 cuts give cheaper solution -> it must be involved in the at most c cuts solution
                            elif c == 0 or CDP[-1][c][COST] < CDP[-1][c_best][COST]:
                                part_best = TPart({subchain[b] for b in barriers(CDP, c)} | sum_m_barr,
                                                  CDP[-1][c][COST] + sum_m_cost)
                                c_best = c
                        if part_best.cost < DP[n][c].cost:
                            DP[n][c] = part_best
            # Subchain head is part of cpath -> a must cut edge is introduced
            else:
                m_cp = next(m for m in tree.succ[head_part[-1]] if m in cpath)
                c_cache = {}
                # Iterate over all feasible cut solution of the subtree T_m_cp
                for k in range(0, c_max):
                    # If the subtree cannot be partitioned with k cuts -> all k-related subcases are infeasible
                    if DP[m_cp][k].cost == math.inf:
                        continue
                    # If the optimal partition cost for k equals to k+1 -> k+1 subcases will the same or more expensive
                    # For given k -> each calculated c cuts are also calculated for k-1 (k->1-c_max, k+1->2-c_max, ...)
                    if k > 0 and DP[m_cp][k - 1].cost <= DP[m_cp][k].cost:
                        continue
                    # Iterate over all possible cuts on the head_part of the subchain
                    for c_head in reversed(range(0, c_max - k)):
                        # Use previously calculated result
                        if c_head in c_cache:
                            barr, opt_cost = c_cache[c_head]
                        else:
                            L_head = sum(tree.nodes[v][RUNTIME] for v in head_part) + c_head * delay
                            barr, opt_cost, _ = partition(runtime, memory, rate, M, N, L_head, 0, len(head_part) - 1,
                                                          delay, unit)
                            # If subchain cannot be partitioned with L_head -> stricter L_head is also infeasible
                            if not barr:
                                break
                            # Precalculate stricter solutions based on the distance between optimal cut and L_head cut
                            for _c in reversed(range(len(barr) - 1, c_head + 1)):
                                c_cache[_c] = (barr, opt_cost)
                        c = k + c_head + 1
                        if (sum_cost := opt_cost + DP[m_cp][k].cost + sum_m_cost) < DP[n][c].cost:
                            DP[n][c] = TPart({subchain[b] for b in barr}.union(DP[m_cp][k].barr, sum_m_barr), sum_cost)
        # If no feasible solution exists for the subtree T_n wrt. L (c_0, c_max = inf) -> no feasible solution for T
        if min(DP[n][0], DP[n][-1]) == math.inf:
            return INFEASIBLE
    c_opt = min(range(len(DP[root])), key=lambda _c: DP[root][_c].cost)
    best_barrs, best_cost = DP[root][c_opt]
    return list(best_barrs) if only_barr else recreate_subchain_blocks(tree, best_barrs), best_cost, c_opt



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from .baseline import baseline_singleton_partitioning, baseline_no_partitioning
from .csp import ibuild_gen_csp_dag, csp_tree_partitioning, csp_gen_tree_partitioning, extract_grp_from_path
from .greedy import min_weight_greedy_partitioning, min_weight_partition_heuristic
from .min_cut import min_weight_chain_decomposition, min_weight_ksplit_clustering, min_weight_tree_clustering



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import networkx as nx
import pandas as pd
from matplotlib import pyplot as plt

from slambuc.gen.cluster.job_tree import (DEF_TASK_CSV, DEF_TASK_CSV_COLS, DEF_TASK_CSV_HEADER, convert_tasks_to_dag,
                                          igenerate_job_tree, DEF_BATCH_CSV)
from slambuc.gen.transform import faasify_dag_by_duplication
from slambuc.misc.plot import draw_tree


def plot_job_dist(task_file: str = DEF_TASK_CSV, min_size: int = 0):
    job_df = pd.read_csv(task_file, usecols=DEF_TASK_CSV_COLS, names=DEF_TASK_CSV_HEADER)
    jobs = job_df.groupby("job")["task"].count()
    jobs = jobs[jobs >= min_size]
    max_size = jobs.max()
    print("Max size:", max_size)
    jobs.plot.hist(bins=50)
    plt.grid(linestyle='dotted', zorder=0)
    plt.show()


def verify_job_tree(job_name: str, task_file: str = DEF_TASK_CSV, draw_weights: bool = False):
    """Generate one job with given *job_name* into service tree and draw tree"""
    print(f"Read data from {task_file}...")
    job_df = pd.read_csv(task_file, usecols=DEF_TASK_CSV_COLS, names=DEF_TASK_CSV_HEADER)
    print(f"Filter task data for job {job_name}...")
    task_data = job_df[job_df["job"] == job_name]
    print(f"Generate tree from {len(task_data)} tasks...")
    dag, root = convert_tasks_to_dag(job_name, task_data)
    tree = faasify_dag_by_duplication(dag, root)
    print(f"Generated tree: {tree}")
    draw_tree(tree, draw_weights=draw_weights)


def check_job_trees(task_file: str = DEF_TASK_CSV, min_size: int = 0, max_size: int = None):
    """Validate generated job trees with size between *min_size* and *max_size*"""
    print(f"Load data from {task_file}...")
    job_df = pd.read_csv(task_file, usecols=DEF_TASK_CSV_COLS, names=DEF_TASK_CSV_HEADER)
    jobs = job_df.groupby("job")["task"].count()
    max_size = max_size if max_size else max(jobs)
    viable_jobs = jobs[(min_size <= jobs) & (jobs <= max_size)]
    print(f"Found {len(viable_jobs)} jobs with size in ({min_size} - {max_size})")
    trees = [t for t in igenerate_job_tree(job_df, min_size=min_size)]
    print("Generated service trees:")
    for tree in trees:
        print(tree, "is tree:", nx.is_tree(tree))


if __name__ == '__main__':
    # plot_job_dist(min_size=30)
    verify_job_tree(job_name="j_905", task_file=DEF_BATCH_CSV, draw_weights=True)
    #
    # check_job_trees(min_size=20, max_size=30)
    # check_job_trees(min_size=30)



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import collections
import math
import multiprocessing
import operator

import networkx as nx

from slambuc.alg import INFEASIBLE, T_RESULTS
from slambuc.alg.service import *
from slambuc.alg.tree.ser.pseudo import SubBTreePart, SubLTreePart, OPT
from slambuc.alg.tree.ser.pseudo_mp import isubtree_splits
from slambuc.alg.util import (ipostorder_dfs, ibacktrack_chain, recreate_subtree_blocks, ileft_right_dfs,
                              par_inst_count, verify_limits)


def _par_ltree_partitioning(ready: multiprocessing.SimpleQueue, sync: dict[int, multiprocessing.SimpleQueue],
                            tree: nx.DiGraph, root: int = 1, M: int = math.inf, L: int = math.inf,
                            N: int = 1, cpath: set[int] = frozenset(), delay: int = 1,
                            bidirectional: bool = True) -> None | dict[int, SubBTreePart]:
    """
    Calculates minimal-cost partitioning of a subgraph with *root* node using the left-right tree traversal approach
    while waits for subcases at *sync* edges.

    This function is designed for running in a separate detached subprocess and synchronizing subresults via
    *SimpleQueue* objects as an IPC method.

    :param ready:           object for signaling the end of partitioning
    :param sync:            object regarding subtrees which results need to be waited for
    :param tree:            service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param root:            root node of the graph
    :param M:               upper memory bound of the partition blocks in MB
    :param L:               latency limit defined on the critical path in ms
    :param N:               available CPU core count
    :param cpath:           critical path nodes
    :param delay:           invocation delay between blocks
    :param bidirectional:   use bidirectional subcase elimination (may introduce quadratic increase in the worst case)
    :return:                tuple of optimal partitioning, reached sum cost and latency on the critical path
    """
    # Allocate empty dict for local subcases combined with prior subcase results
    TDP = {}
    # Process subtrees of T in a bottom-up traversal order
    for p, n in ipostorder_dfs(tree, root):
        # Init empty data structure for optimal subcases T_n[v,b]
        DP = collections.defaultdict(lambda: collections.defaultdict(dict))
        # Store sync point of the overlapped subtree
        sentinel = None
        r_n, d_n, t_n, m_n = tree[p][n][RATE], tree[p][n][DATA], tree.nodes[n][RUNTIME], tree.nodes[n][MEMORY]
        # Traverse subtree T_n in a left-right traversal
        #       |--(j-1.)--> [v_p]   |--(i-1.)--> [b_p]                                 | left side
        # -->  [u]  ----(j.)---->  [v]  ----(i.)---->  [b]  ----(max_k.)----> [b_k]     | right side
        for e_uvp, b_p, e_vb in ileft_right_dfs(tree, n):
            v, b = e_vb
            # Merge edge u -> v as consider node v and all its first 0 child (b=0)
            if b == 0:
                # Cache passed sync point
                if v in sync:
                    sentinel = sync[v]
                # SINGLETON: C[n,0] - calculate the default subcase of singleton partition of root node n
                if v == n:
                    cost = r_n * (d_n + t_n) + sum(par_inst_count(r_n, ns[RATE], N) * ns[DATA]
                                                   for ns in tree.succ[n].values())
                    if n in cpath:
                        nc = next(filter(lambda c: c in cpath, tree.successors(n)), None)
                        # n_v = 1 for root node n; add caching overhead on critical path here
                        lat = d_n + t_n + (math.ceil(tree[n][nc][RATE] / (r_n * N)) * tree[n][nc][DATA] if nc else 0)
                    else:
                        lat = 0
                    DP[e_vb][lat][m_n, m_n] = SubLTreePart(cost, 1, {n})
                # MERGE: C[v,0] - subcase of merging single node v into the n-rooted block containing node u
                else:
                    u, v_p = e_uvp
                    r_v, d_v = tree[u][v][RATE], tree[u][v][DATA]
                    t_v, m_v = tree.nodes[v][RUNTIME], tree.nodes[v][MEMORY]
                    # v_p = 0 => no prior sibling node, C[v,0] = C[u,j-1] + delta(u,v)
                    for lat_u in DP[e_uvp]:
                        for mem_u in DP[e_uvp][lat_u]:
                            sub_u = DP[e_uvp][lat_u][mem_u]
                            # Calculate memory prefetch and operative memory demands separately
                            mem = mem_u[0] + m_v, max(mem_u[1], min(math.ceil(r_v / r_n), N) * m_v)
                            if max(mem) > M:
                                # Infeasible subcase due to exceeded memory constraint
                                continue
                            # v in cpath => u in cpath
                            if v in cpath:
                                rel_r_v = math.ceil(r_v / (tree[next(tree.predecessors(u))][u][RATE] * N))
                                n_v = rel_r_v * sub_u.mul
                                vc = next(filter(lambda c: c in cpath, tree.successors(v)), None)
                                w_v = math.ceil(tree[v][vc][RATE] / (r_v * N)) * tree[v][vc][DATA] if vc else 0
                                # Recalculate caching overhead (remove prior block's write out from E2E latency)
                                if (lat := lat_u + n_v * (t_v - d_v + w_v)) > L:
                                    # Infeasible subcase due to exceeded latency constraint
                                    continue
                            else:
                                lat, n_v = lat_u, sub_u.mul
                            # Calculate sum cost by adding merged node runtime and caching
                            cost = (sub_u.cost + par_inst_count(r_n, r_v, N) * (t_v - d_v) +
                                    sum(par_inst_count(r_n, vs[RATE], N) * vs[DATA] for vs in tree.succ[v].values()))
                            # Check whether the current subcase is a better sub-solution for state (e_vb, lat, mem)
                            if lat not in DP[e_vb] or mem not in DP[e_vb][lat] or cost < DP[e_vb][lat][mem].cost:
                                # Eliminate prior subcases that are dominated by the new subcase
                                for l, m in tuple((_l, _m) for _l in DP[e_vb] if _l >= lat
                                                  for _m in DP[e_vb][_l] if (_m[0] >= mem[0] and _m[0] >= mem[0]
                                                                             and DP[e_vb][_l][_m].cost >= cost)):
                                    if len(DP[e_vb][l]) > 1:
                                        del DP[e_vb][l][m]
                                    else:
                                        del DP[e_vb][l]
                                # Add superior subcase
                                DP[e_vb][lat][mem] = SubLTreePart(cost, n_v, sub_u.barr)
            # CUT: C[v,b]
            else:
                # b_k <- C[v_i, d(v_i)]
                b_k = collections.deque(tree.successors(b), 1).pop() if len(tree.succ[b]) else 0
                # Previously merged edge: v -> b  =>  [root ... -> u -> v -> b ...]   |   [b]  --(max_k.)--> [b_w]
                DP[e_vb] = DP[b, b_k]
                # Cut edge: v -> b  |  [root ... -> u -> v] - [b ...]   |   v --(i-1.)--> b_p
                for lat_v in DP[v, b_p]:
                    for mem_v in DP[v, b_p][lat_v]:
                        # Waiting for dependent subprocess to be finished and pull optimal subcases
                        if b not in TDP:
                            TDP.update(sentinel.get())
                            sentinel.close()
                        for lat_b, opt_b in TDP[b].items():
                            # b in cpath => v in cpath
                            if (lat := lat_v + delay + lat_b if b in cpath else lat_v) > L:
                                # Infeasible subcase due to exceeded latency constraint
                                continue
                            sub_v = DP[v, b_p][lat_v][mem_v]
                            # Only concatenate blocks
                            cost, barr = sub_v.cost + opt_b.cost, sub_v.barr | opt_b.barr
                            # Check whether the current subcase is a better sub-solution for state (e_vb, lat, mem)
                            if lat not in DP[e_vb] or mem_v not in DP[e_vb][lat] or cost < DP[e_vb][lat][mem_v].cost:
                                if bidirectional:
                                    # Eliminate prior subcases that are dominated by the new subcase
                                    for l, m in tuple((_l, _m) for _l in DP[e_vb] if _l >= lat
                                                      for _m in DP[e_vb][_l] if (_m[0] >= mem_v[0] and _m[0] >= mem_v[0]
                                                                                 and DP[e_vb][_l][_m].cost >= cost)):
                                        if len(DP[e_vb][l]) > 1:
                                            del DP[e_vb][l][m]
                                        else:
                                            del DP[e_vb][l]
                                # Add superior subcase
                                DP[e_vb][lat][mem_v] = SubLTreePart(cost, sub_v.mul, barr)
        # Cache the best subcase for subtree T_n
        n_w = collections.deque(tree.successors(n), 1).pop() if len(tree.succ[n]) else 0
        # Store the best subcases of subtree T_n
        TDP[n] = {lat_n: min(dp.values(), key=operator.itemgetter(OPT)) for lat_n, dp in DP[n, n_w].items()}
    # Notify waiting process and push optimal subcases or return TDP locally for the main thread
    return ready.put(TDP) if ready else TDP


def pseudo_par_mp_ltree_partitioning(tree: nx.DiGraph, root: int = 1, M: int = math.inf, L: int = math.inf,
                                     N: int = 1, cp_end: int = None, delay: int = 1,
                                     bidirectional: bool = True) -> T_RESULTS:
    """
    Calculates minimal-cost partitioning of a service graph(tree) with respect to an upper bound **M** on the total
    memory of blocks and a latency constraint **L** defined on the subchain between *root* and *cp_end* nodes.

    Partitioning is calculated using the left-right tree traversal approach.

    Arbitrary disjoint subtrees are partitioned in separate subprocesses.

    Block metrics are calculated based on parallelized execution platform model.

    :param tree:            service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param root:            root node of the graph
    :param M:               upper memory bound of the partition blocks in MB
    :param L:               latency limit defined on the critical path in ms
    :param N:               available CPU core count
    :param cp_end:          tail node of the critical path in the form of subchain[root -> cp_end]
    :param delay:           invocation delay between blocks
    :param bidirectional:   use bidirectional subcase elimination (may introduce quadratic increase in the worst case)
    :return:                tuple of optimal partitioning, reached sum cost and latency on the critical path
    """
    # Critical path
    cpath = set(ibacktrack_chain(tree, root, cp_end))
    # Verify the min values of limits for a feasible solution
    if not all(verify_limits(tree, cpath, M, L)):
        # No feasible solution due to too strict limits
        return INFEASIBLE
    workers, events, sync = [], {}, None
    try:
        # Split tree and initiate subprocesses to calculate subcases
        for (p, v), sync in isubtree_splits(tree, root):
            # Collect event of sync points
            sync = {s: events[s] for s in sync}
            if p is not PLATFORM:
                # Add own event for signaling readiness
                ready = events.setdefault(v, multiprocessing.SimpleQueue())
                # Create and start worker process
                workers.append(multiprocessing.Process(target=_par_ltree_partitioning, name=f"subtree_{v}",
                                                       args=(ready, sync, tree, v, M, L, N, cpath, delay,
                                                             bidirectional), daemon=False))
                workers[-1].start()
        # Process last/topmost subtree in the main process and get subresults locally
        TDP = _par_ltree_partitioning(None, sync, tree, root, M, L, N, cpath, delay, bidirectional)
    except BaseException:
        # Terminate all initiated subprocesses in case of interruption
        for w in workers:
            w.terminate()
        # Reraise exception for further handling
        raise
    finally:
        # Wait for all subprocesses to terminate for closing all used resources
        for w in workers:
            w.join(timeout=0)
            if w.is_alive():
                w.kill()
                w.join()
    # Subcases under the root node contain the feasible partitioning
    if opt_lat := min(TDP[root], key=lambda _l: TDP[root][_l].cost, default=None):
        opt = TDP[root][opt_lat]
        return recreate_subtree_blocks(tree, opt.barr), opt.cost, opt_lat
    else:
        # No feasible solution
        return INFEASIBLE



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math

import networkx as nx

from slambuc.alg import INFEASIBLE, T_RESULTS
from slambuc.alg.tree import seq_tree_partitioning
from slambuc.alg.util import recalculate_partitioning
from slambuc.gen.transform import transform_autonomous_caching


def cacheless_path_tree_partitioning(tree: nx.DiGraph, root: int = 1, M: int = math.inf, N: int = 1, L: int = math.inf,
                                     cp_end: int = None, delay: int = 1, valid: bool = True) -> T_RESULTS:
    """
    Calculates minimal-cost partitioning using *seq_tree_partitioning* without considering data externalization.

    :param tree:    service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param root:    root node of the graph
    :param M:       upper memory bound of the partition blocks (in MB)
    :param N:       upper CPU core bound of the partition blocks
    :param L:       latency limit defined on the critical path (in ms)
    :param cp_end:  tail node of the critical path in the form of subchain[root -> cp_end]
    :param delay:   invocation delay between blocks
    :param valid:   return only L-feasible solutions
    :return:        tuple of optimal partition, sum cost of the partitioning, and optimal number of cuts
    """
    partition, *_ = seq_tree_partitioning(tree, root, M, N, L, cp_end, delay, unit=1)
    if not partition:
        return INFEASIBLE
    sum_cost, sum_lat = recalculate_partitioning(tree, partition, root, N, cp_end, delay)
    return INFEASIBLE if valid and sum_lat > L else (partition, sum_cost, sum_lat)


def stateful_path_tree_partitioning(tree: nx.DiGraph, root: int = 1, M: int = math.inf, N: int = 1, L: int = math.inf,
                                    cp_end: int = None, delay: int = 1) -> T_RESULTS:
    """
    Calculates minimal-cost partitioning using *seq_tree_partitioning* while considering data implicit state
    externalization.

    Input tree is preprocessed and function runtimes are altered to incorporate data read/write overheads.

    :param tree:    service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param root:    root node of the graph
    :param M:       upper memory bound of the partition blocks (in MB)
    :param N:       upper CPU core bound of the partition blocks
    :param L:       latency limit defined on the critical path (in ms)
    :param cp_end:  tail node of the critical path in the form of subchain[root -> cp_end]
    :param delay:   invocation delay between blocks
    :return:        tuple of optimal partition, sum cost of the partitioning, and optimal number of cuts
    """
    partition, *_ = seq_tree_partitioning(transform_autonomous_caching(tree, root, copy=True),
                                          root, M, N, L, cp_end, delay, unit=1)
    return (partition, *recalculate_partitioning(tree, partition, root, N, cp_end, delay)) if partition else INFEASIBLE



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections.abc import Generator

from slambuc.alg.service import Flavor

# Function signature types
T_BLOCK = list[int]  # list of block nodes
T_PART = list[T_BLOCK]  # partitioning as list of blocks
T_RESULTS = tuple[T_PART, int, int]  # Partitioning, sum cost, sum latency
T_PART_GEN = Generator[T_PART]
T_IBLOCK = list[int, int]  # Block interval start and end nodes
T_IBLOCK_GEN = Generator[T_IBLOCK]
T_FBLOCK = tuple[T_BLOCK, Flavor]
T_FPART = list[T_FBLOCK]
T_FRESULTS = tuple[T_FPART, int, int]
T_BARRS = list[int] | set[int]  # list/set of barrier nodes
T_BARRS_GEN = Generator[T_BARRS]
T_BRESULTS = tuple[T_BARRS, int, int]

# Constants for attribute indices in DP matrix
import math

BARR, COST, LAT = 0, 1, 2
# Constant for block cache index
MEM, CPU = 0, 3
# Constant for ILP model names
LP_LAT = 'C_LAT'

# Infeasible solution [partitioning, opt_cost, opt_lat]
INFEASIBLE = ([], math.inf, None)



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import collections
import itertools
import math
import operator
import typing

import networkx as nx

from slambuc.alg import INFEASIBLE, T_RESULTS
from slambuc.alg.service import *
from slambuc.alg.service.common import WEIGHT
from slambuc.alg.util import (ipostorder_dfs, ileft_right_dfs, ibacktrack_chain, recreate_subtree_blocks,
                              recalculate_ser_partitioning, ipostorder_edges, verify_limits)

# Constants for attribute index
OPT = 0


class WeightedSubBTreePart(typing.NamedTuple):
    """Store subtree partitioning attributes for a given edge-weighted subcase."""
    weight: int = 0  # Cumulative weights of covered edges in the subtree partitioning
    barr: set[int] = set()  # Barrier/heading nodes of the given subtree partitioning

    def __repr__(self):
        return repr(tuple(self))


def biheuristic_btree_partitioning(tree: nx.DiGraph, root: int = 1, M: int = math.inf, L: int = math.inf,
                                   cp_end: int = None, delay: int = 1, Epsilon: float = 0.0, Lambda: float = 0.0,
                                   bidirectional: bool = True) -> T_RESULTS:
    """
    Calculates minimal-cost partitioning of a service graph(tree) with respect to an upper bound **M** on the total
    memory of blocks and a latency constraint **L** defined on the subchain between *root* and *cp_end* nodes, while
    applying the bottom-up tree traversal approach.

    Cost approximation ratio *Epsilon* controls the maximum deviation from the cost-optimal partitioning
    (Epsilon=0.0 enforces the algorithm to calculate exact solution) in exchange for reduces subcase calculations.

    Latency approximation ratio (*Lambda*) controls the maximum deviation with respect to the latency limit $L$
    (Lambda=0.0 enforces no rounding) in exchange for reduces subcase calculations.
    
    Block metrics are calculated based on serialized execution platform model.

    Provide suboptimal partitioning due to the simplified and inaccurate latency rounding.

    :param tree:            service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param root:            root node of the graph
    :param M:               upper memory bound of the partition blocks in MB
    :param L:               latency limit defined on the critical path in ms
    :param cp_end:          tail node of the critical path in the form of subchain[root -> cp_end]
    :param delay:           invocation delay between blocks
    :param Epsilon:         weight factor for state space trimming (0 <= Eps < 1, Eps = 0 falls back to exact calc.)
    :param Lambda:          latency factor for state space trimming (0 <= Lambda, Lambda = 0 falls back to exact calc.)
    :param bidirectional:   use bidirectional subcase elimination (may introduce quadratic increase in the worst case)
    :return:                tuple of optimal partitioning, reached sum cost and latency on the critical path
    """
    # Set of critical path's nodes
    cpath = set(ibacktrack_chain(tree, root, cp_end))
    # Verify the min values of limits for a feasible solution
    if not all(verify_limits(tree, cpath, M, L)):
        # No feasible solution due to too strict limits
        return INFEASIBLE
    # Init empty data structure for optimal results of subtrees T_v
    DP = {_v: collections.defaultdict(dict) for _v in tree if _v is not PLATFORM}
    # Number of nodes
    t_size = len(DP)
    # Iterate nodes in a bottom-up traversal order
    for p, v in ipostorder_dfs(tree, root):
        # SINGLETON: calculate the default subcase of singleton partition of node v
        r_v, d_v, t_v, m_v = tree[p][v][RATE], tree[p][v][DATA], tree.nodes[v][RUNTIME], tree.nodes[v][MEMORY]
        # Only add fetching overhead for root node and omit caching to ensure lat monotonicity of upward merging
        lat = d_v + t_v if v == root else t_v if v in cpath else 0
        # For singleton blocks weight is 0 by default
        DP[v][lat, lat][m_v] = WeightedSubBTreePart(0, {v})
        w_max = 0
        # Bottom-up propagation for considering v's descendant subcases in sequential order (v is not leaf)
        # -->   p   ---->   v   ----[i.]---->   b
        for b in tree.successors(v):
            # Init empty data structure of subcase T[v,b]
            _cache = collections.defaultdict(dict)
            r_b, d_b, t_b = tree[v][b][RATE], tree[v][b][DATA], tree.nodes[b][RUNTIME]
            # Calculate possible latency/memory combinations while dropping subcases of v and b from DP
            for ((lat_v, blk_lat_v), DPv), ((lat_b, blk_lat_b), DPb) in itertools.product(DP[v].items(),
                                                                                          DP.pop(b).items()):
                for (mem_v, sub_v), (mem_b, sub_b) in itertools.product(DPv.items(), DPb.items()):
                    # Latency calculation in case v -> b edge is in cpath
                    if b in cpath:
                        # CUT: v -> b edge is marked as a cut (opt subcases of b have the dedicated memory value 0)
                        if mem_b == OPT:
                            # Add caching overhead of inter-block call
                            top_blk_lat = blk_lat_v + math.ceil(r_b / r_v) * d_b
                            # Add invocation delay and fetch overhead of inter-block call
                            if (lat := top_blk_lat + delay + d_b + lat_b) > L:
                                # Infeasible subcase due to exceeded latency constraint
                                continue
                        # MERGE: v -> b edge is marked as a merge
                        else:
                            # Top block's latency comes for only the single cp node v, hence n_v = 1
                            top_blk_lat = blk_lat_v + math.ceil(r_b / r_v) * blk_lat_b
                            # Calculate new latency based on the recalculated top block's latency
                            if (lat := lat_b - blk_lat_b + top_blk_lat) > L:
                                # Infeasible subcase due to exceeded latency constraint
                                continue
                    else:
                        # Top block's attributes remain the same, either v in cpath or not
                        top_blk_lat, lat = blk_lat_v, lat_v
                    # CUT: v -> b edge is marked as a cut
                    if mem_b == OPT:
                        # Sub-partitions are just concatenated
                        mem, weight, barr = mem_v, sub_v.weight + sub_b.weight, sub_v.barr | sub_b.barr
                    # MERGE: v -> b edge is marked as a merge
                    elif (mem := mem_v + mem_b) > M:
                        # Infeasible subcase due to exceeded memory constraint
                        continue
                    else:
                        # Top blocks of the two sub-partitions are merged together with root node v
                        weight, barr = sub_v.weight + sub_b.weight + r_b * d_b, sub_v.barr | sub_b.barr - {b}
                    # Store the max-weight subcase
                    sub_lats = lat, top_blk_lat
                    # Check whether the current subcase is a better sub-solution for state (v, sub_lats, mem)
                    if mem not in _cache[sub_lats] or weight > _cache[sub_lats][mem].weight:
                        if bidirectional:
                            # Eliminate prior subcases that are dominated by the new subcase
                            for l, m in tuple((_l, _m) for _l in _cache if _l[0] >= lat and _l[1] >= top_blk_lat
                                              for _m in _cache[_l] if _m >= mem and _cache[_l][_m].weight <= weight):
                                if len(_cache[l]) > 1:
                                    del _cache[l][m]
                                else:
                                    del _cache[l]
                        # Add superior subcase
                        _cache[sub_lats][mem] = WeightedSubBTreePart(weight, barr)
                        w_max = max(w_max, weight)
            # Store max subcases as C(v,i-1) for the next iteration of the propagation process
            DP[v] = _cache
        # Trim state space based on given bi-criteria parameters
        if len(tree.succ[v]) > 0:
            # Calculate trim interval sizes
            l_max = L if L < math.inf else max((_l[0] for _l in DP[v]), default=0)
            l_scale = ((Lambda * l_max) / t_size) if Lambda > 0.0 and l_max > 0 else 1.0
            w_scale = ((Epsilon * w_max) / t_size) if Epsilon > 0.0 and w_max > 0 else 1.0
            # print(f"w int size: {w_scale}, l int size: {l_scale}")
            _kept_subcases = {}
            # Search for dominated subcases
            for (lat, top_blk_lat), dpv in tuple(DP[v].items()):
                for mem, subcase in tuple(dpv.items()):
                    # Track states as a tuple of interval sequence numbers
                    if (lw_state := (math.ceil(lat / l_scale),
                                     math.ceil(top_blk_lat / l_scale),
                                     math.ceil(subcase.weight / w_scale))) in _kept_subcases:
                        kept_lats, kept_mem = _kept_subcases[lw_state]
                        # Evaluate dominance relation between visited and current subcase
                        if kept_mem > mem:
                            # Drop dominated subcase
                            del DP[v][kept_lats][kept_mem]
                            if not len(DP[v][kept_lats]):
                                del DP[v][kept_lats]
                        else:
                            continue
                    # Add subcase as kept state
                    _kept_subcases[lw_state] = (lat, top_blk_lat), mem
        # Store the weight-opt subcases wrt. latency for node v encoded with memory value 0
        for sub_v in DP[v].values():
            sub_v[OPT] = max(sub_v.values(), key=operator.itemgetter(0))
    # Subcases under the root node contain the feasible partitioning
    if opt_lats := max(DP[root], key=lambda _l: DP[root][_l][OPT].weight, default=None):
        opt = DP[root][opt_lats][OPT]
        return recreate_subtree_blocks(tree, opt.barr), opt.weight, opt_lats[0]
    else:
        # No feasible solution
        return INFEASIBLE


def biheuristic_tree_partitioning(tree: nx.DiGraph, root: int = 1, M: int = math.inf, L: int = math.inf,
                                  cp_end: int = None, delay: int = 1, Epsilon: float = 0.0, Lambda: float = 0.0,
                                  bidirectional: bool = True) -> T_RESULTS:
    """
    Calculates minimal-cost partitioning of a service graph(tree) with respect to an upper bound **M** on the total
    memory of blocks and a latency constraint **L** defined on the subchain between *root* and *cp_end* nodes, while
    applying the bottom-up tree traversal approach.

    Provide suboptimal partitioning due to the simplified and inaccurate latency rounding.

    Recalculates original sum cost and latency metrics.

    :param tree:            service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param root:            root node of the graph
    :param M:               upper memory bound of the partition blocks in MB
    :param L:               latency limit defined on the critical path in ms
    :param cp_end:          tail node of the critical path in the form of subchain[root -> cp_end]
    :param delay:           invocation delay between blocks
    :param Epsilon:         weight factor for state space trimming (0 <= Eps < 1, Eps = 0 falls back to exact calc.)
    :param Lambda:          latency factor for state space trimming (0 <= Lambda, Lambda = 0 falls back to exact calc.)
    :param bidirectional:   use bidirectional subcase elimination (may introduce quadratic increase in the worst case)
    :return:                tuple of optimal partitioning, reached sum cost and latency on the critical path
    """
    partition, *_ = biheuristic_btree_partitioning(tree, root, M, L, cp_end, delay, Epsilon, Lambda, bidirectional)
    if partition:
        return partition, *recalculate_ser_partitioning(tree, partition, root, cp_end, delay)
    else:
        # No feasible solution
        return INFEASIBLE


########################################################################################################################


class WeightedSubLTreePart(typing.NamedTuple):
    """Store subtree partitioning attributes for a given edge-weighted subcase."""
    weight: int = 0  # Cumulative weights of covered edges in the subtree partitioning
    top_lat: int = 0  # Calculated latency for the topmost partition block
    mul: int = 1  # Last serialization multiplier of the top/first block of the subtree partitioning
    barr: set[int] = set()  # Barrier/heading nodes of the given subtree partitioning

    def __repr__(self):
        return repr(tuple(self))


def bifptas_ltree_partitioning(tree: nx.DiGraph, root: int = 1, M: int = math.inf, L: int = math.inf,
                               cp_end: int = None, delay: int = 1, Epsilon: float = 0.0, Lambda: float = 0.0,
                               bidirectional: bool = True) -> T_RESULTS:
    """
    Calculates minimal-cost partitioning of a service graph(tree) with respect to an upper bound **M** on the total
    memory of blocks and a latency constraint **L** defined on the subchain between *root* and *cp_end* nodes, while
    applying the left-right tree traversal approach.

    Cost approximation ratio *Epsilon* controls the maximum deviation from the cost-optimal partitioning
    (Epsilon=0.0 enforces the algorithm to calculate exact solution) in exchange for reduces subcase calculations.

    Latency violation ratio (*Lambda*) controls the maximum violating deviation from the latency limit $L$
    (Lambda=0.0 enforces no violation)  in exchange for reduces subcase calculations.

    Block metrics are calculated based on serialized execution platform model.

    :param tree:            service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param root:            root node of the graph
    :param M:               upper memory bound of the partition blocks in MB
    :param L:               latency limit defined on the critical path in ms
    :param cp_end:          tail node of the critical path in the form of subchain[root -> cp_end]
    :param delay:           invocation delay between blocks
    :param Epsilon:         weight factor for state space trimming (0 <= Eps < 1, Eps = 0 falls back to exact calc.)
    :param Lambda:          latency factor for state space trimming (0 <= Lambda, Lambda = 0 falls back to exact calc.)
    :param bidirectional:   use bidirectional subcase elimination (may introduce quadratic increase in the worst case)
    :return:                tuple of optimal partitioning, reached sum cost and latency on the critical path
    """
    # Set of critical path's nodes
    cpath = set(ibacktrack_chain(tree, root, cp_end))
    # Verify the min values of limits for a feasible solution
    if not all(verify_limits(tree, cpath, M, L)):
        # No feasible solution due to too strict limits
        return INFEASIBLE
    # Init empty data structure for optimal results of subtrees
    TDP = {n: {} for n in tree if n is not PLATFORM}
    t_size = len(TDP)
    # Calculate latency rounding interval and scaled latency upper bound
    if Lambda > 0.0 and L < math.inf:
        l_scale = ((Lambda * L) / t_size)
        L_hat = math.floor(L / l_scale) + t_size
    else:
        l_scale, L_hat = 1.0, L
    # Process subtrees of T in a bottom-up traversal order
    for p, n in ipostorder_dfs(tree, root):
        # Init empty data structure for optimal subcases T_n[v,b]
        DP = collections.defaultdict(lambda: collections.defaultdict(dict))
        # Traverse subtree T_n in a left-right traversal
        #       |--(j-1.)--> [v_p]   |--(i-1.)--> [b_p]                                 | left side
        # -->  [u]  ----(j.)---->  [v]  ----(i.)---->  [b]  ----(max_k.)----> [b_k]     | right side
        w_max = 0
        for e_uvp, b_p, e_vb in ileft_right_dfs(tree, n):
            v, b = e_vb
            # Merge edge u -> v
            if b == 0:
                # SINGLETON: C[n,0] - calculate the default subcase of singleton partition of root node n
                if v == n:
                    r_n, d_n, t_n = tree[p][n][RATE], tree[p][n][DATA], tree.nodes[n][RUNTIME]
                    if n in cpath:
                        nc = next(filter(lambda c: c in cpath, tree.successors(n)), None)
                        lat_n = d_n + t_n + (math.ceil(tree[n][nc][RATE] / r_n) * tree[n][nc][DATA] if nc else 0)
                    else:
                        lat_n = 0
                    # Rounding latency value for singleton block
                    lat = math.ceil(lat_n / l_scale)
                    # For singleton blocks weight is 0 by default
                    DP[e_vb][lat][tree.nodes[n][MEMORY]] = WeightedSubLTreePart(0, lat_n, 1, {n})
                # MERGE: C[v,0] - subcase of merging single node v into the n-rooted block containing node u
                else:
                    u, v_p = e_uvp
                    r_v, d_v, t_v = tree[u][v][RATE], tree[u][v][DATA], tree.nodes[v][RUNTIME]
                    # v_p = 0 => no prior sibling node, C[v,0] = C[u,j-1] + delta(u,v)
                    for lat_u in DP[e_uvp]:
                        for mem_u in DP[e_uvp][lat_u]:
                            if (mem := mem_u + tree.nodes[v][MEMORY]) > M:
                                # Infeasible subcase due to exceeded memory constraint
                                continue
                            sub_u = DP[e_uvp][lat_u][mem_u]
                            # v in cpath => u in cpath
                            if v in cpath:
                                n_v = math.ceil(r_v / tree[next(tree.predecessors(u))][u][RATE]) * sub_u.mul
                                vc = next(filter(lambda c: c in cpath, tree.successors(v)), None)
                                o_v = math.ceil(tree[v][vc][RATE] / r_v) * tree[v][vc][DATA] if vc else 0
                                # Calculate latency of merged node v and add rounded value to top block
                                top_lat = sub_u.top_lat + n_v * (t_v - d_v + o_v)
                                if (lat := math.ceil(top_lat / l_scale)) > L_hat:
                                    # Infeasible subcase due to exceeded latency constraint
                                    continue
                            else:
                                # Top block's latency attributes remains the same
                                lat, top_lat, n_v = lat_u, sub_u.top_lat, sub_u.mul
                            # Store subcase with weight increased by the merged/covered edge weight
                            weight = sub_u.weight + r_v * d_v
                            DP[e_vb][lat][mem] = WeightedSubLTreePart(weight, top_lat, n_v, sub_u.barr)
            # CUT: C[v,b]
            else:
                # b_k <- C[v_i, d(v_i)]
                b_k = collections.deque(tree.successors(b), 1).pop() if len(tree.succ[b]) else 0
                # Previously merged edge: v -> b  =>  [root ... -> u -> v -> b ...]   |   [b]  --(max_k.)--> [b_w]
                DP[e_vb] = DP[b, b_k]
                # Cut edge: v -> b  |  [root ... -> u -> v] - [b ...]   |   v --(i-1.)--> b_p
                for lat_v in DP[v, b_p]:
                    for mem_v in DP[v, b_p][lat_v]:
                        for lat_b, opt_b in TDP[b].items():
                            # add delay for cut and round b in cpath => v in cpath
                            # lat_b_hat = math.ceil((delay + opt_b.lat) / l_scale) if b in cpath else 0
                            lat_b_hat = (lat_b - math.ceil(opt_b.top_lat / l_scale)
                                         + math.ceil((opt_b.top_lat + delay) / l_scale)) if b in cpath else 0
                            if (lat := lat_v + lat_b_hat) > L_hat:
                                # Infeasible subcase due to exceeded latency constraint
                                continue
                            sub_v = DP[v, b_p][lat_v][mem_v]
                            # Sub-partitions are just concatenated
                            top_lat, weight, barr = sub_v.top_lat, sub_v.weight + opt_b.weight, sub_v.barr | opt_b.barr
                            # Check whether the current subcase is a better sub-solution for state (e_vb, lat, mem)
                            if (lat not in DP[e_vb] or mem_v not in DP[e_vb][lat]
                                    or weight > DP[e_vb][lat][mem_v].weight):
                                if bidirectional:
                                    # Eliminate prior subcases that are dominated by the new subcase
                                    for l, m in tuple((_l, _m) for _l in DP[e_vb] if _l >= lat for _m in DP[e_vb][_l]
                                                      if _m >= mem_v and DP[e_vb][_l][_m].weight <= weight):
                                        if len(DP[e_vb][l]) > 1:
                                            del DP[e_vb][l][m]
                                        else:
                                            del DP[e_vb][l]
                                # Add superior subcase
                                DP[e_vb][lat][mem_v] = WeightedSubLTreePart(weight, top_lat, sub_v.mul, barr)
                                # Accumulate max calculated weight for subcases of n
                                w_max = max(w_max, weight)
        # Cache the best subcase for subtree T_n getting from the last successor of node n
        sn_last = collections.deque(tree.successors(n), 1).pop() if len(tree.succ[n]) else 0
        # Trim state space based on given bi-criteria parameters
        if len(tree.succ[n]) > 0:
            # Calculate trim interval size for weights
            w_scale = ((Epsilon * w_max) / t_size) if Epsilon > 0.0 and w_max > 0 else 1.0
            _kept_subcases = {}
            # Search for dominated subcases
            for lat, dpn in tuple(DP[n, sn_last].items()):
                for mem, subcase in tuple(dpn.items()):
                    # Track states as a tuple of interval sequence numbers
                    if (lw_state := (lat, math.ceil(subcase.weight / w_scale))) in _kept_subcases:
                        kept_mem = _kept_subcases[lw_state]
                        # Evaluate dominance relation between visited and current subcase
                        if kept_mem > mem:
                            # Drop dominated subcase
                            if len(DP[n, sn_last][lat]) > 1:
                                del DP[n, sn_last][lat][kept_mem]
                            else:
                                del DP[n, sn_last][lat]
                        else:
                            continue
                    # Add subcase as kept state
                    _kept_subcases[lw_state] = mem
        # Store the max-weight subcases of subtree T_n
        for lat_n, dp in DP[n, sn_last].items():
            TDP[n][lat_n] = max(dp.values(), key=operator.itemgetter(0))
    # Subcases under the root node contain the feasible partitioning
    if opt_lat := max(TDP[root], key=lambda _l: TDP[root][_l].weight, default=None):
        opt = TDP[root][opt_lat]
        return recreate_subtree_blocks(tree, opt.barr), opt.weight, opt_lat
    else:
        # No feasible solution
        return INFEASIBLE


def bifptas_tree_partitioning(tree: nx.DiGraph, root: int = 1, M: int = math.inf, L: int = math.inf,
                              cp_end: int = None, delay: int = 1, Epsilon: float = 0.0, Lambda: float = 0.0,
                              bidirectional: bool = True) -> T_RESULTS:
    """
    Calculates minimal-cost partitioning of a service graph(tree) with respect to an upper bound **M** on the total
    memory of blocks and a latency constraint **L** defined on the subchain between *root* and *cp_end* nodes, while
    applying the left-right tree traversal approach.

    Recalculates original sum cost and latency metrics.

    :param tree:            service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param root:            root node of the graph
    :param M:               upper memory bound of the partition blocks in MB
    :param L:               latency limit defined on the critical path in ms
    :param cp_end:          tail node of the critical path in the form of subchain[root -> cp_end]
    :param delay:           invocation delay between blocks
    :param Epsilon:         weight factor for state space trimming (0 <= Eps < 1, Eps = 0 falls back to exact calc.)
    :param Lambda:          latency factor for state space trimming (0 <= Lambda, Lambda = 0 falls back to exact calc.)
    :param bidirectional:   use bidirectional subcase elimination (may introduce quadratic increase in the worst case)
    :return:                tuple of optimal partitioning, reached sum cost and latency on the critical path
    """
    partition, *_ = bifptas_ltree_partitioning(tree, root, M, L, cp_end, delay, Epsilon, Lambda, bidirectional)
    if partition:
        return partition, *recalculate_ser_partitioning(tree, partition, root, cp_end, delay)
    else:
        # No feasible solution
        return INFEASIBLE


########################################################################################################################


class WeightedDualSubLTreePart(typing.NamedTuple):
    """Store subtree partitioning attributes for a given edge-weighted subcase."""
    mem: int = 0  # Memory demand of the topmost block in the subtree partitioning
    top_lat: int = 0  # Calculated latency for the topmost partition block
    mul: int = 1  # Last serialization multiplier of the top/first block of the subtree partitioning
    barr: set[int] = set()  # Barrier/heading nodes of the given subtree partitioning

    def __repr__(self):
        return repr(tuple(self))


def bifptas_dual_ltree_partitioning(tree: nx.DiGraph, root: int = 1, M: int = math.inf, L: int = math.inf,
                                    cp_end: int = None, delay: int = 1, Epsilon: float = 0.0, Lambda: float = 0.0,
                                    bidirectional: bool = True) -> T_RESULTS:
    """
    Calculates minimal-cost partitioning of a service graph(tree) with respect to an upper bound **M** on the total
    memory of blocks and a latency constraint **L** defined on the subchain between *root* and *cp_end* nodes, while
    applying the left-right tree traversal approach.

    Cost approximation ratio *Epsilon* controls the maximum deviation from the cost-optimal partitioning
    (Epsilon=0.0 enforces the algorithm to calculate exact solution) in exchange for reduces subcase calculations.

    Latency violation ratio (*Lambda*) controls the maximum violating deviation from the latency limit $L$
    (Lambda=0.0 enforces no violation)  in exchange for reduces subcase calculations.

    Instead of direct cost calculations, the cumulative overheads of externalized states are subject to minimization
    as a different formalization of the same optimization problem.

    Block metrics are calculated based on serialized execution platform model.

    :param tree:            service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param root:            root node of the graph
    :param M:               upper memory bound of the partition blocks in MB
    :param L:               latency limit defined on the critical path in ms
    :param cp_end:          tail node of the critical path in the form of subchain[root -> cp_end]
    :param delay:           invocation delay between blocks
    :param Epsilon:         weight factor for state space trimming (0 <= Eps < 1, Eps = 0 falls back to exact calc.)
    :param Lambda:          latency factor for state space trimming (0 <= Lambda, Lambda = 0 falls back to exact calc.)
    :param bidirectional:   use bidirectional subcase elimination (may introduce quadratic increase in the worst case)
    :return:                tuple of optimal partitioning, reached sum cost and latency on the critical path
    """
    # Set of critical path's nodes
    cpath = set(ibacktrack_chain(tree, root, cp_end))
    # Verify the min values of limits for a feasible solution
    if not all(verify_limits(tree, cpath, M, L)):
        # No feasible solution due to too strict limits
        return INFEASIBLE
    # Init empty data structure for optimal results of subtrees
    TDP = {n: {} for n in tree if n is not PLATFORM}
    t_size = len(TDP)
    # Calculate latency rounding interval and scaled latency upper bound
    if Lambda > 0.0 and L < math.inf:
        l_scale = ((Lambda * L) / t_size)
        L_hat = math.floor(L / l_scale) + t_size
    else:
        l_scale, L_hat = 1.0, L
    w_max = max(d[WEIGHT] for u, v, d in ipostorder_edges(tree, root, data=True)
                if d.setdefault(WEIGHT, d[RATE] * d[DATA]) and tree.nodes[u][MEMORY] + tree.nodes[v][MEMORY] <= M)
    w_scale = ((Epsilon * w_max) / t_size) if Epsilon > 0.0 and w_max > 0 else 1.0
    # Process subtrees of T in a bottom-up traversal order
    for p, n in ipostorder_dfs(tree, root):
        # Init empty data structure for optimal subcases T_n[v,b]
        DP = collections.defaultdict(lambda: collections.defaultdict(dict))
        # Traverse subtree T_n in a left-right traversal
        #       |--(j-1.)--> [v_p]   |--(i-1.)--> [b_p]                                 | left side
        # -->  [u]  ----(j.)---->  [v]  ----(i.)---->  [b]  ----(max_k.)----> [b_k]     | right side
        for e_uvp, b_p, e_vb in ileft_right_dfs(tree, n):
            v, b = e_vb
            # Merge edge u -> v
            if b == 0:
                # SINGLETON: C[n,0] - calculate the default subcase of singleton partition of root node n
                if v == n:
                    t_n, m_n = tree.nodes[n][RUNTIME], tree.nodes[n][MEMORY]
                    r_n, d_n = tree[p][n][RATE], tree[p][n][DATA]
                    if n in cpath:
                        nc = next(filter(lambda c: c in cpath, tree.successors(n)), None)
                        lat_n = d_n + t_n + (math.ceil(tree[n][nc][RATE] / r_n) * tree[n][nc][DATA] if nc else 0)
                    else:
                        lat_n = 0
                    # Rounding latency value for singleton block
                    lat = math.ceil(lat_n / l_scale)
                    # For singleton blocks weight is 0 by default
                    DP[e_vb][lat][0] = WeightedDualSubLTreePart(m_n, lat_n, 1, {n})
                # MERGE: C[v,0] - subcase of merging single node v into the n-rooted block containing node u
                else:
                    u, _ = e_uvp
                    t_v, m_v = tree.nodes[v][RUNTIME], tree.nodes[v][MEMORY]
                    r_v, d_v, w_v = tree[u][v][RATE], tree[u][v][DATA], tree[u][v][WEIGHT]
                    # v_p = 0 => no prior sibling node, C[v,0] = C[u,j-1] + delta(u,v)
                    for lat_u in DP[e_uvp]:
                        for weight_u in DP[e_uvp][lat_u]:
                            sub_u = DP[e_uvp][lat_u][weight_u]
                            if (mem := sub_u.mem + m_v) > M:
                                # Infeasible subcase due to exceeded memory constraint
                                continue
                            # v in cpath => u in cpath
                            if v in cpath:
                                n_v = math.ceil(r_v / tree[next(tree.predecessors(u))][u][RATE]) * sub_u.mul
                                vc = next(filter(lambda c: c in cpath, tree.successors(v)), None)
                                o_v = math.ceil(tree[v][vc][RATE] / r_v) * tree[v][vc][DATA] if vc else 0
                                # Calculate latency of merged node v and add rounded value to top block
                                top_lat = sub_u.top_lat + n_v * (t_v - d_v + o_v)
                                if (lat := math.ceil(top_lat / l_scale)) > L_hat:
                                    # Infeasible subcase due to exceeded latency constraint
                                    continue
                            else:
                                # Top block's latency attributes remains the same
                                lat, top_lat, n_v = lat_u, sub_u.top_lat, sub_u.mul
                            # Store subcase with weight increased by the merged/covered edge weight
                            weight = weight_u + math.ceil(w_v / w_scale)
                            DP[e_vb][lat][weight] = WeightedDualSubLTreePart(mem, top_lat, n_v, sub_u.barr)
            # CUT: C[v,b]
            else:
                # b_k <- C[v_i, d(v_i)]
                b_k = collections.deque(tree.successors(b), 1).pop() if len(tree.succ[b]) else 0
                # Previously merged edge: v -> b  =>  [root ... -> u -> v -> b ...]   |   [b]  --(max_k.)--> [b_w]
                DP[e_vb] = DP[b, b_k]
                # Cut edge: v -> b  |  [root ... -> u -> v] - [b ...]   |   v --(i-1.)--> b_p
                for lat_v in DP[v, b_p]:
                    for weight_v in DP[v, b_p][lat_v]:
                        sub_v = DP[v, b_p][lat_v][weight_v]
                        for (weight_b, lat_b), opt_b in TDP[b].items():
                            # add delay for cut and round b in cpath => v in cpath
                            lat_b_hat = (lat_b - math.ceil(opt_b.top_lat / l_scale)
                                         + math.ceil((opt_b.top_lat + delay) / l_scale)) if b in cpath else 0
                            if (lat := lat_v + lat_b_hat) > L_hat:
                                # Infeasible subcase due to exceeded latency constraint
                                continue
                            # Sub-partitions are just concatenated
                            top_lat, mem = sub_v.top_lat, sub_v.mem
                            weight, barr = weight_v + weight_b, sub_v.barr | opt_b.barr
                            # Check whether the current subcase is a better sub-solution for state (e_vb, lat, mem)
                            if lat not in DP[e_vb] or weight not in DP[e_vb][lat] or mem < DP[e_vb][lat][weight].mem:
                                if bidirectional:
                                    # Eliminate prior subcases that are dominated by the new subcase
                                    for l, w in tuple((_l, _w) for _l in DP[e_vb] if _l >= lat for _w in DP[e_vb][_l]
                                                      if _w <= weight and DP[e_vb][_l][_w].mem >= mem):
                                        if len(DP[e_vb][l]) > 1:
                                            del DP[e_vb][l][w]
                                        else:
                                            del DP[e_vb][l]
                                # Add superior subcase
                                DP[e_vb][lat][weight] = WeightedDualSubLTreePart(mem, top_lat, sub_v.mul, barr)
        # Cache the best subcase for subtree T_n getting from the last successor of node n
        sn_last = collections.deque(tree.successors(n), 1).pop() if len(tree.succ[n]) else 0
        # Store the max-weight subcases of subtree T_n
        for lat_n, dp in DP[n, sn_last].items():
            weight_n_max = max(dp)
            TDP[n][weight_n_max, lat_n] = dp[weight_n_max]
    # Subcases under the root node contain the feasible partitioning
    if opt_wl := max(TDP[root], key=operator.itemgetter(0), default=None):
        return recreate_subtree_blocks(tree, TDP[root][opt_wl].barr), *opt_wl
    else:
        # No feasible solution
        return INFEASIBLE


def bifptas_dual_tree_partitioning(tree: nx.DiGraph, root: int = 1, M: int = math.inf, L: int = math.inf,
                                   cp_end: int = None, delay: int = 1, Epsilon: float = 0.0, Lambda: float = 0.0,
                                   bidirectional: bool = True) -> T_RESULTS:
    """
    Calculates minimal-cost partitioning of a service graph(tree) with respect to an upper bound **M** on the total
    memory of blocks and a latency constraint **L** defined on the subchain between *root* and *cp_end* nodes, while
    applying a different formalization of the optimal partitioning problem.

    Cost approximation ratio *Epsilon* controls the maximum deviation from the cost-optimal partitioning
    (Epsilon=0.0 enforces the algorithm to calculate exact solution) in exchange for reduces subcase calculations.

    Latency violation ratio (*Lambda*) controls the maximum violating deviation from the latency limit $L$
    (Lambda=0.0 enforces no violation)  in exchange for reduces subcase calculations.

    Block metrics are calculated based on serialized execution platform model.

    Recalculates original sum cost and latency metrics.

    :param tree:            service graph annotated with node runtime(ms), memory(MB) and edge rates and data overheads(ms)
    :param root:            root node of the graph
    :param M:               upper memory bound of the partition blocks in MB
    :param L:               latency limit defined on the critical path in ms
    :param cp_end:          tail node of the critical path in the form of subchain[root -> cp_end]
    :param delay:           invocation delay between blocks
    :param Epsilon:         weight factor for state space trimming (0 <= Eps < 1, Eps = 0 falls back to exact calc.)
    :param Lambda:          latency factor for state space trimming (0 <= Lambda, Lambda = 0 falls back to exact calc.)
    :param bidirectional:   use bidirectional subcase elimination (may introduce quadratic increase in the worst case)
    :return:                tuple of optimal partitioning, reached sum cost and latency on the critical path
    """
    partition, *_ = bifptas_dual_ltree_partitioning(tree, root, M, L, cp_end, delay, Epsilon, Lambda, bidirectional)
    if partition:
        return partition, *recalculate_ser_partitioning(tree, partition, root, cp_end, delay)
    else:
        # No feasible solution
        return INFEASIBLE



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import networkx as nx

from slambuc.alg import T_RESULTS
from slambuc.alg.service import PLATFORM
from slambuc.alg.util import recalculate_partitioning


def baseline_singleton_partitioning(tree: nx.DiGraph, root: int = 1, N: int = 1, cp_end: int = None,
                                    delay: int = 1, **kwargs) -> T_RESULTS:
    """
    Derive the trivial partitioning of grouping all nodes into one single block.

    :param tree:    service graph annotated with node runtime(ms), memory(MB) and edge rate and data
    :param root:    root node of the graph
    :param N:       available CPU core count
    :param cp_end:  tail node of the critical path in the form of subchain[root -> cp_end]
    :param delay:   invocation delay between blocks
    :return:        tuple of partitioning, reached sum cost and latency on the critical path
    """
    partitioning = [sorted(filter(lambda v: v is not PLATFORM, tree))]
    return partitioning, *recalculate_partitioning(tree, partitioning, root, N, cp_end, delay)


def baseline_no_partitioning(tree: nx.DiGraph, root: int = 1, N: int = 1, cp_end: int = None,
                             delay: int = 1, **kwargs) -> T_RESULTS:
    """
    Derive the trivial solution of not merging any of the given tree nodes.

    :param tree:    service graph annotated with node runtime(ms), memory(MB) and edge rate and data
    :param root:    root node of the graph
    :param N:       available CPU core count
    :param cp_end:  tail node of the critical path in the form of subchain[root -> cp_end]
    :param delay:   invocation delay between blocks
    :return:        tuple of partitioning, reached sum cost and latency on the critical path
    """
    partitioning = [[v] for v in tree if v is not PLATFORM]
    return partitioning, *recalculate_partitioning(tree, partitioning, root, N, cp_end, delay)



# Copyright 2023 Janos Czentye
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import itertools
import math

from slambuc.alg import MEM, COST, LAT, CPU, T_BARRS, T_BRESULTS
from slambuc.alg.chain.dp.mtx import State


def min_chain_partitioning(runtime: list[int], memory: list[int], rate: list[int], M: int = math.inf,
                           N: int = math.inf, L: int = math.inf, start: int = 0, end: int = None,
                           delay: int = 1, unit: int = 100) -> T_BRESULTS:
    """
    Calculates minimal-cost partitioning of a chain based on the node properties of *running time*, *memory usage* and
    *invocation rate* with respect to an upper bound **M** on the total memory of blocks and a latency constraint **L**
    defined on the subchain between *start* and *end* nodes.

    Cost calculation relies on the rounding *unit* and number of vCPU cores *N*, whereas platform invocation *delay*
    is used for latency calculations.

    It gives optimal result only in case the cost function regarding the chain attributes is sub-additive,
    that is k_opt = k_min is guaranteed for each case.

    Instead of full partitioning it only returns the list of barrier nodes.

    :param runtime: running times in ms
    :param memory:  memory requirements in MB
    :param rate:    avg. rate of function invocations
    :param M:       upper memory bound of the partition blocks (in MB)
    :param N:       upper CPU core bound of the partition blocks
    :param L:       latency limit defined on the critical path in the form of subchain[start -> end] (in ms)
    :param delay:   invocation delay between blocks
    :param start:   head node of the latency-limited subchain
    :param end:     tail node of the latency-limited subchain
    :param unit:    rounding unit for the cost calculation (default: 100 ms)
    :return:        tuple of barrier nodes, sum cost of the partitioning, and the calculated latency on the subchain
    """
    n = len(runtime)
    end = end if end is not None else n - 1

    def block_memory(_v: int, cumsum: list = (0,)) -> int:
        """Calculate memory of a block from the cached cumulative sum of block[v+1, w] or block[b, v-1]."""
        cumsum[MEM] += memory[_v]
        return cumsum[MEM]

    def block_cpu(_b: int, cumsum: list = (0, 0, 0, (0, 1))) -> int:
        """Calculate CPU need of a block from the cached cumulative sum of block[b+1, w]."""
        cumsum[CPU][0] = max(cumsum[CPU][0], rate[_b])
        cumsum[CPU][1] = max(cumsum[CPU][1], math.ceil(cumsum[CPU][0] / rate[_b]))
        return cumsum[CPU][1]

    def block_cost(_b: int, cumsum: list = (0, 0)) -> int:
        """Calculate running time of block[b, w] from the cached cumulative sum of block[b+1, w]."""
        cumsum[COST] += runtime[_b]
        return rate[_b] * (math.ceil(cumsum[COST] / unit) * unit)

    def block_latency(_b: int, _w: int, cumsum: list = (0, 0, 0)) -> int:
        """Calculate relevant latency for block[b, w] from the cached cumulative sum of block[b+1, w]."""
        # Do not consider latency if no intersection
        if end < _b or _w < start:
            return 0
        if b < start:
            return cumsum[LAT]
        cumsum[LAT] += runtime[_b]
        # Ignore delay if latency path starts within the subchain
        return delay + cumsum[LAT] if start < _b else cumsum[LAT]

    # Check lower bound for latency limit
    if L < (lat_min := sum(runtime[start: end + 1])):
        return None, None, lat_min
    # Check if memory constraint allows feasible solutions for the given latency constraint
    k_min = max(math.ceil(sum(memory[start: end + 1]) / M),
                sum(1 for i, j in itertools.pairwise(rate) if math.ceil(j / i) > N))
    k_max = math.floor(min((L - sum(runtime[start: end + 1])) / delay + 1, n))
    if k_max < k_min:
        return None, None, None
    # Check single node partitioning
    if len(runtime) == 1:
        return [0], block_cost(0), block_latency(0, 0)
    # Initialize DP matrix with an additional trailing element for the backward reference of the trivial singleton cases
    DP = [State() for _ in range(n + 1)]
    # Initialize default reference values for the trivial singleton subcases
    DP[-1] = State(None, 0, 0)
    # Calculate Dynamic Programming matrix
    for w in range(0, n):
        # Cache for the cumulative sum based calculation of expanding block's memory, runtime and latency values
        _cache = [0, 0, 0, [0, 1]]
        for b in reversed(range(0, w + 1)):
            # Larger groups will not satisfy the memory constraint M
            if block_memory(b, _cache) > M or block_cpu(b, _cache) > N:
                break
            # For the singleton subcases (b=0) the reference values are zero: DP[-1] -> 0
            if (lat := DP[b - 1].lat + block_latency(b, w, _cache)) <= L:
                # Store and overwrite subcases with equal costs (<=) to consider larger blocks for lower latency
                if (cost := DP[b - 1].cost + block_cost(b, _cache)) <= DP[w].cost:
                    DP[w] = State(b, cost, lat)
    return (extract_min_barr(DP), DP[n - 1].cost, DP[n - 1].lat) if DP[n - 1].cost < math.inf else DP[n - 1]


def extract_min_barr(DP: list[State]) -> T_BARRS:
    """
    Extract barrier nodes form DP list by iteratively backtracking minimal cost subcases.

    :param DP:  dynamic programming structure storing intermediate *States*
    :return:    list of barrier nodes
    """
    barr = [DP[-2].barr]
    while barr[-1]:
        barr.append(DP[barr[-1] - 1].barr)
    barr.reverse()
    return barr
